"""Building blocks for TiTok.

Copyright (2024) Bytedance Ltd. and/or its affiliates

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 

Reference: 
    https://github.com/mlfoundations/open_clip/blob/main/src/open_clip/transformer.py
    https://github.com/baofff/U-ViT/blob/main/libs/timm.py
"""

import math
import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint
from collections import OrderedDict
import einops
from einops.layers.torch import Rearrange
import torch.nn.functional as F
from einops import rearrange
from accelerate.utils.operations import gather
from torch.amp import autocast
from typing import Mapping, Text, Tuple
import numpy as np
from medlat.modules.in_and_out import PatchEmbed, ToPixel


def modulate(x, shift, scale):
    return x * (1 + scale) + shift


class ResidualAttentionBlock(nn.Module):
    def __init__(
            self,
            d_model,
            n_head,
            mlp_ratio = 4.0,
            act_layer = nn.GELU,
            norm_layer = nn.LayerNorm
        ):
        super().__init__()

        self.ln_1 = norm_layer(d_model)
        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.mlp_ratio = mlp_ratio
        # optionally we can disable the FFN
        if mlp_ratio > 0:
            self.ln_2 = norm_layer(d_model)
            mlp_width = int(d_model * mlp_ratio)
            self.mlp = nn.Sequential(OrderedDict([
                ("c_fc", nn.Linear(d_model, mlp_width)),
                ("gelu", act_layer()),
                ("c_proj", nn.Linear(mlp_width, d_model))
            ]))

    def attention(
            self,
            x: torch.Tensor
    ):
        return self.attn(x, x, x, need_weights=False)[0]

    def forward(
            self,
            x: torch.Tensor,
    ):
        attn_output = self.attention(x=self.ln_1(x))
        x = x + attn_output
        if self.mlp_ratio > 0:
            x = x + self.mlp(self.ln_2(x))
        return x

if hasattr(torch.nn.functional, 'scaled_dot_product_attention'):
    ATTENTION_MODE = 'flash'
else:
    try:
        import xformers
        import xformers.ops
        ATTENTION_MODE = 'xformers'
    except:
        ATTENTION_MODE = 'math'
print(f'attention mode is {ATTENTION_MODE}')


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, L, C = x.shape

        qkv = self.qkv(x)
        if ATTENTION_MODE == 'flash':
            qkv = einops.rearrange(qkv, 'B L (K H D) -> K B H L D', K=3, H=self.num_heads).float()
            q, k, v = qkv[0], qkv[1], qkv[2]  # B H L D
            x = torch.nn.functional.scaled_dot_product_attention(q, k, v)
            x = einops.rearrange(x, 'B H L D -> B L (H D)')
        elif ATTENTION_MODE == 'xformers':
            qkv = einops.rearrange(qkv, 'B L (K H D) -> K B L H D', K=3, H=self.num_heads)
            q, k, v = qkv[0], qkv[1], qkv[2]  # B L H D
            x = xformers.ops.memory_efficient_attention(q, k, v)
            x = einops.rearrange(x, 'B L H D -> B L (H D)', H=self.num_heads)
        elif ATTENTION_MODE == 'math':
            qkv = einops.rearrange(qkv, 'B L (K H D) -> K B H L D', K=3, H=self.num_heads)
            q, k, v = qkv[0], qkv[1], qkv[2]  # B H L D
            attn = (q @ k.transpose(-2, -1)) * self.scale
            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)
            x = (attn @ v).transpose(1, 2).reshape(B, L, C)
        else:
            raise NotImplemented

        x = self.proj(x)
        x = self.proj_drop(x)
        return x


def drop_path(x, drop_prob: float = 0., training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class UViTBlock(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, skip=False, use_checkpoint=False):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        self.skip_linear = nn.Linear(2 * dim, dim) if skip else None
        self.use_checkpoint = use_checkpoint

    def forward(self, x, skip=None):
        if self.use_checkpoint:
            return torch.utils.checkpoint.checkpoint(self._forward, x, skip)
        else:
            return self._forward(x, skip)

    def _forward(self, x, skip=None):
        if self.skip_linear is not None:
            x = self.skip_linear(torch.cat([x, skip], dim=-1))
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x
    

def _expand_token(token, batch_size: int):
    return token.unsqueeze(0).expand(batch_size, -1, -1)


class TiTokEncoder(nn.Module):
    def __init__(self, img_size, in_channels=3, patch_size=16, hidden_size=768, depth=12, num_heads=12, num_latent_tokens=64, token_size=12, quantize_mode="vq", is_legacy=True, mlp_ratio=4.0):
        super().__init__()
        self.img_size = img_size
        self.patch_size = patch_size
        self.hidden_size = hidden_size
        self.depth = depth
        self.num_heads = num_heads
        self.num_latent_tokens = num_latent_tokens
        self.token_size = token_size

        # Determine spatial dimensions (2D or 3D)
        if isinstance(self.img_size, int):
            self.dims = 2
            self.img_size = (self.img_size, self.img_size)
        elif isinstance(self.img_size, (list, tuple)):
            self.dims = len(self.img_size)
            self.img_size = tuple(self.img_size)
        else:
            raise ValueError(f"Unknown img_size type: {type(self.img_size)}")

        if isinstance(self.patch_size, int):
            self.patch_size = (self.patch_size,) * self.dims
        elif isinstance(self.patch_size, (list, tuple)):
            self.patch_size = tuple(self.patch_size)
            if len(self.patch_size) != self.dims:
                raise ValueError(f"patch_size dims {len(self.patch_size)} do not match image dims {self.dims}")
        else:
            raise ValueError(f"Unknown patch_size type: {type(self.patch_size)}")

        # Compute grid size (tuple for each spatial dim)
        self.grid_size = tuple(img // p for img, p in zip(self.img_size, self.patch_size))
        self.n_patches = int(np.prod(self.grid_size))

        if quantize_mode == "vae":
            self.token_size = self.token_size * 2  # needs to split into mean and std

        self.is_legacy = is_legacy

        self.patch_embed = PatchEmbed(to_embed='conv', img_size=self.img_size, patch_size=self.patch_size, in_chans=in_channels, embed_dim=self.hidden_size)

        scale = self.hidden_size ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(1, self.hidden_size))
        self.positional_embedding = nn.Parameter(
            scale * torch.randn(self.n_patches + 1, self.hidden_size)
        )
        self.latent_token_positional_embedding = nn.Parameter(
            scale * torch.randn(self.num_latent_tokens, self.hidden_size)
        )
        self.ln_pre = nn.LayerNorm(self.hidden_size)
        self.transformer = nn.ModuleList()
        for i in range(self.depth):
            self.transformer.append(ResidualAttentionBlock(
                self.hidden_size, self.num_heads, mlp_ratio=mlp_ratio
            ))
        self.ln_post = nn.LayerNorm(self.hidden_size)

        # Use appropriate ConvNd for projection according to dims
        if self.dims == 2:
            self.conv_out = nn.Conv2d(self.hidden_size, self.token_size, kernel_size=1, bias=True)
        elif self.dims == 3:
            self.conv_out = nn.Conv3d(self.hidden_size, self.token_size, kernel_size=1, bias=True)
        else:
            raise NotImplementedError("Only 2D and 3D data supported.")

    def forward(self, pixel_values, latent_tokens):
        batch_size = pixel_values.shape[0]
        x = pixel_values
        x = self.patch_embed(x)
        x = torch.cat([_expand_token(self.class_embedding, x.shape[0]).to(x.dtype), x], dim=1)
        x = x + self.positional_embedding.to(x.dtype)  # (B, N_patch+1, C)

        latent_tokens = _expand_token(latent_tokens, x.shape[0]).to(x.dtype)
        latent_tokens = latent_tokens + self.latent_token_positional_embedding.to(x.dtype)
        x = torch.cat([x, latent_tokens], dim=1)

        x = self.ln_pre(x)
        x = x.permute(1, 0, 2)  # (L, B, C)
        for i in range(self.depth):
            x = self.transformer[i](x)
        x = x.permute(1, 0, 2)  # (B, L, C)

        latent_tokens = x[:, 1 + self.n_patches :]
        latent_tokens = self.ln_post(latent_tokens)
        # "fake" ND shape: project transformer outputs to spatial + token axes for conv
        if self.is_legacy:
            if self.dims == 2:
                # (B, C, num_latent_tokens, 1)
                latent_tokens = latent_tokens.reshape(batch_size, self.hidden_size, self.num_latent_tokens, 1)
            elif self.dims == 3:
                # (B, C, num_latent_tokens, 1, 1)
                latent_tokens = latent_tokens.reshape(batch_size, self.hidden_size, self.num_latent_tokens, 1, 1)
        else:
            # New shape: (B, num_latent_tokens, C, 1) or (B, num_latent_tokens, C, 1, 1) --> permute to (B, C, num_latent_tokens, 1) or (B, C, num_latent_tokens, 1, 1)
            if self.dims == 2:
                latent_tokens = latent_tokens.reshape(batch_size, self.num_latent_tokens, self.hidden_size, 1).permute(0, 2, 1, 3)
            elif self.dims == 3:
                latent_tokens = latent_tokens.reshape(batch_size, self.num_latent_tokens, self.hidden_size, 1, 1).permute(0, 2, 1, 3, 4)
        latent_tokens = self.conv_out(latent_tokens)

        # Final reshape: flatten "extra" dims (beyond batch and token_size), keep last dim as num_latent_tokens
        # e.g., for 2d: (B, token_size, 1, num_latent_tokens)  for 3d: (B, token_size, 1, 1, num_latent_tokens)
        out_shape = (batch_size, self.token_size) + (1,) * (self.dims - 1) + (self.num_latent_tokens,)
        latent_tokens = latent_tokens.reshape(out_shape)
        return latent_tokens
    

class TiTokDecoder(nn.Module):
    def __init__(
        self, 
        img_size, 
        codebook_size,
        out_channels=3, 
        patch_size=16, 
        hidden_size=768, 
        depth=12, 
        num_heads=12, 
        num_latent_tokens=64, 
        token_size=12, 
        quantize_mode="vq", 
        is_legacy=True, 
        mlp_ratio=4.0
    ):
        super().__init__()
        # Set up image and patch sizes and determine spatial dims (2d vs 3d)
        if isinstance(img_size, int):
            self.img_size = (img_size, img_size)
        else:
            self.img_size = tuple(img_size)
        if isinstance(patch_size, int):
            self.patch_size = (patch_size,) * len(self.img_size)
        else:
            self.patch_size = tuple(patch_size)
        self.dims = len(self.img_size)
        assert self.dims in [2, 3], "Only 2D and 3D supported"

        # Compute grid size for each spatial dim
        self.grid_size_tuple = tuple([i // p for i, p in zip(self.img_size, self.patch_size)])
        self.grid_size = self.grid_size_tuple  # tuple
        self.n_grid = int(np.prod(self.grid_size_tuple))

        self.hidden_size = hidden_size
        self.depth = depth
        self.num_heads = num_heads
        self.num_latent_tokens = num_latent_tokens
        self.token_size = token_size
        self.codebook_size = codebook_size
        if quantize_mode == "vae":
            self.token_size = self.token_size * 2 # needs to split into mean and std
        self.is_legacy = is_legacy
        self.mlp_ratio = mlp_ratio

        self.decoder_embed = nn.Linear(
            self.token_size, self.hidden_size, bias=True)
        scale = self.hidden_size ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(1, self.hidden_size))
        self.positional_embedding = nn.Parameter(
            scale * torch.randn(self.n_grid + 1, self.hidden_size)
        )
        self.mask_token = nn.Parameter(scale * torch.randn(1, 1, self.hidden_size))
        self.latent_token_positional_embedding = nn.Parameter(
            scale * torch.randn(self.num_latent_tokens, self.hidden_size)
        )
        self.ln_pre = nn.LayerNorm(self.hidden_size)
        self.transformer = nn.ModuleList(
            [ResidualAttentionBlock(self.hidden_size, self.num_heads, mlp_ratio=mlp_ratio)
            for _ in range(self.depth)]
        )
        self.ln_post = nn.LayerNorm(self.hidden_size)

        # Adapt FFN and out for 2D/3D
        if self.is_legacy:
            if self.dims == 2:
                self.ffn = nn.Sequential(
                    nn.Conv2d(self.hidden_size, 2 * self.hidden_size, 1, padding=0, bias=True),
                    nn.Tanh(),
                    nn.Conv2d(2 * self.hidden_size, self.codebook_size, 1, padding=0, bias=True),
                )
                self.conv_out = nn.Identity()
            elif self.dims == 3:
                self.ffn = nn.Sequential(
                    nn.Conv3d(self.hidden_size, 2 * self.hidden_size, 1, padding=0, bias=True),
                    nn.Tanh(),
                    nn.Conv3d(2 * self.hidden_size, self.codebook_size, 1, padding=0, bias=True),
                )
                self.conv_out = nn.Identity()
        else:
            # Directly predicting RGB pixels for both 2D and 3D
            if self.dims == 2:
                self.ffn = nn.Sequential(
                    nn.Conv2d(self.hidden_size, np.prod(self.patch_size) * 3, 1, padding=0, bias=True),
                    Rearrange('b (p1 p2 c) h w -> b c (h p1) (w p2)',
                        p1 = self.patch_size[0], p2 = self.patch_size[1]),
                )
                self.conv_out = nn.Conv2d(3, 3, 3, padding=1, bias=True)
            elif self.dims == 3:
                self.ffn = nn.Sequential(
                    nn.Conv3d(self.hidden_size, np.prod(self.patch_size) * out_channels, 1, padding=0, bias=True),  # Typically only 1 channel for 3D, but can be adapted
                    Rearrange('b (p1 p2 p3 c) d h w -> b c (d p1) (h p2) (w p3)', 
                        p1 = self.patch_size[0], p2 = self.patch_size[1], p3 = self.patch_size[2], c=1),
                )
                self.conv_out = nn.Conv3d(1, 1, 3, padding=1, bias=True)  # Adapt as needed

    def forward(self, z_quantized):
        # Accept 2D or 3D z_quantized based on self.dims and num_latent_tokens
        N = z_quantized.shape[0]
        if self.dims == 2:
            # z_quantized: (N, C, 1, num_latent_tokens)
            C, H, W = z_quantized.shape[1:]
            assert H == 1 and W == self.num_latent_tokens, f"{H}, {W}, {self.num_latent_tokens}"
            x = z_quantized.reshape(N, C * H, W).permute(0, 2, 1) # NLD
            grid_shape = self.grid_size
        elif self.dims == 3:
            # z_quantized: (N, C, 1, 1, num_latent_tokens)
            C, H, W, D = z_quantized.shape[1:]
            assert H == 1 and W == 1 and D == self.num_latent_tokens, f"{H}, {W}, {D}, {self.num_latent_tokens}"
            x = z_quantized.reshape(N, C * H * W, D).permute(0, 2, 1) # NLD (N, num_latent_tokens, C')
            grid_shape = self.grid_size
        else:
            raise NotImplementedError("Only 2D and 3D supported for TiTokDecoder forward.")

        x = self.decoder_embed(x)  # (N, seq_len, hidden_size)
        batchsize, seq_len, _ = x.shape

        mask_tokens = self.mask_token.repeat(batchsize, self.n_grid, 1).to(x.dtype)
        mask_tokens = torch.cat(
            [_expand_token(self.class_embedding, mask_tokens.shape[0]).to(mask_tokens.dtype), mask_tokens], dim=1
        )
        mask_tokens = mask_tokens + self.positional_embedding.to(mask_tokens.dtype)
        x = x + self.latent_token_positional_embedding[:seq_len]
        x = torch.cat([mask_tokens, x], dim=1)
        
        x = self.ln_pre(x)
        x = x.permute(1, 0, 2)  # NLD -> LND
        for i in range(self.depth):
            x = self.transformer[i](x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = x[:, 1:1+self.n_grid] # remove cls embed
        x = self.ln_post(x)
        
        if self.dims == 2:
            # N L D -> N D H W
            x = x.permute(0, 2, 1).reshape(batchsize, self.hidden_size, grid_shape[0], grid_shape[1])
        elif self.dims == 3:
            # N L D -> N D D H W (flattened L to 3D)
            x = x.permute(0, 2, 1).reshape(batchsize, self.hidden_size, grid_shape[0], grid_shape[1], grid_shape[2])
        else:
            raise NotImplementedError("Only 2D and 3D supported for FFN reshape.")

        x = self.ffn(x.contiguous())
        x = self.conv_out(x)
        return x
    
    def get_last_layer(self) -> nn.Parameter:
        if self.is_legacy:
            return self.ffn[-1].weight  # Last Conv2d or Conv3d in legacy mode
        else:
            return self.conv_out.weight  # Conv2d or Conv3d in direct RGB mode


class TATiTokDecoder(TiTokDecoder):
    def __init__(self, config):
        super().__init__(config)
        scale = self.width ** -0.5
        self.text_context_length = config.model.vq_model.get("text_context_length", 77)
        self.text_embed_dim = config.model.vq_model.get("text_embed_dim", 768)
        self.text_guidance_proj = nn.Linear(self.text_embed_dim, self.width)
        self.text_guidance_positional_embedding = nn.Parameter(scale * torch.randn(self.text_context_length, self.width))

    def forward(self, z_quantized, text_guidance):
        N, C, H, W = z_quantized.shape
        assert H == 1 and W == self.num_latent_tokens, f"{H}, {W}, {self.num_latent_tokens}"
        x = z_quantized.reshape(N, C*H, W).permute(0, 2, 1) # NLD
        x = self.decoder_embed(x)

        batchsize, seq_len, _ = x.shape

        mask_tokens = self.mask_token.repeat(batchsize, self.grid_size**2, 1).to(x.dtype)
        mask_tokens = torch.cat([_expand_token(self.class_embedding, mask_tokens.shape[0]).to(mask_tokens.dtype),
                                    mask_tokens], dim=1)
        mask_tokens = mask_tokens + self.positional_embedding.to(mask_tokens.dtype)
        x = x + self.latent_token_positional_embedding[:seq_len]
        x = torch.cat([mask_tokens, x], dim=1)

        text_guidance = self.text_guidance_proj(text_guidance)
        text_guidance = text_guidance + self.text_guidance_positional_embedding
        x = torch.cat([x, text_guidance], dim=1)
        
        x = self.ln_pre(x)
        x = x.permute(1, 0, 2)  # NLD -> LND
        for i in range(self.num_layers):
            x = self.transformer[i](x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = x[:, 1:1+self.grid_size**2] # remove cls embed
        x = self.ln_post(x)
        # N L D -> N D H W
        x = x.permute(0, 2, 1).reshape(batchsize, self.width, self.grid_size, self.grid_size)
        x = self.ffn(x.contiguous())
        x = self.conv_out(x)
        return x
    

class WeightTiedLMHead(nn.Module):
    def __init__(self, embeddings, target_codebook_size):
        super().__init__()
        self.weight = embeddings.weight
        self.target_codebook_size = target_codebook_size

    def forward(self, x):
        # x shape: [batch_size, seq_len, embed_dim]
        # Get the weights for the target codebook size
        weight = self.weight[:self.target_codebook_size]  # Shape: [target_codebook_size, embed_dim]
        # Compute the logits by matrix multiplication
        logits = torch.matmul(x, weight.t())  # Shape: [batch_size, seq_len, target_codebook_size]
        return logits


class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """
    def __init__(self, hidden_size, frequency_embedding_size=256):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """
        # https://github.com/openai/glide-text2im/blob/main/glide_text2im/nn.py
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=t.device)
        args = t[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq)
        return t_emb


class ResBlock(nn.Module):
    """
    A residual block that can optionally change the number of channels.
    :param channels: the number of input channels.
    """

    def __init__(
        self,
        channels
    ):
        super().__init__()
        self.channels = channels

        self.in_ln = nn.LayerNorm(channels, eps=1e-6)
        self.mlp = nn.Sequential(
            nn.Linear(channels, channels, bias=True),
            nn.SiLU(),
            nn.Linear(channels, channels, bias=True),
        )

        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(channels, 3 * channels, bias=True)
        )

    def forward(self, x, y):
        shift_mlp, scale_mlp, gate_mlp = self.adaLN_modulation(y).chunk(3, dim=-1)
        h = modulate(self.in_ln(x), shift_mlp, scale_mlp)
        h = self.mlp(h)
        return x + gate_mlp * h


class FinalLayer(nn.Module):
    """
    The final layer adopted from DiT.
    """
    def __init__(self, model_channels, out_channels):
        super().__init__()
        self.norm_final = nn.LayerNorm(model_channels, elementwise_affine=False, eps=1e-6)
        self.linear = nn.Linear(model_channels, out_channels, bias=True)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(model_channels, 2 * model_channels, bias=True)
        )

    def forward(self, x, c):
        shift, scale = self.adaLN_modulation(c).chunk(2, dim=-1)
        x = modulate(self.norm_final(x), shift, scale)
        x = self.linear(x)
        return x


class SimpleMLPAdaLN(nn.Module):
    """
    The MLP for Diffusion Loss.
    :param in_channels: channels in the input Tensor.
    :param model_channels: base channel count for the model.
    :param out_channels: channels in the output Tensor.
    :param z_channels: channels in the condition.
    :param num_res_blocks: number of residual blocks per downsample.
    """

    def __init__(
        self,
        in_channels,
        model_channels,
        out_channels,
        z_channels,
        num_res_blocks,
        grad_checkpointing=False,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.grad_checkpointing = grad_checkpointing

        self.time_embed = TimestepEmbedder(model_channels)
        self.cond_embed = nn.Linear(z_channels, model_channels)

        self.input_proj = nn.Linear(in_channels, model_channels)

        res_blocks = []
        for i in range(num_res_blocks):
            res_blocks.append(ResBlock(
                model_channels,
            ))

        self.res_blocks = nn.ModuleList(res_blocks)
        self.final_layer = FinalLayer(model_channels, out_channels)

        self.initialize_weights()

    def initialize_weights(self):
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)
        self.apply(_basic_init)

        # Initialize timestep embedding MLP
        nn.init.normal_(self.time_embed.mlp[0].weight, std=0.02)
        nn.init.normal_(self.time_embed.mlp[2].weight, std=0.02)

        # Zero-out adaLN modulation layers
        for block in self.res_blocks:
            nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(block.adaLN_modulation[-1].bias, 0)

        # Zero-out output layers
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].bias, 0)
        nn.init.constant_(self.final_layer.linear.weight, 0)
        nn.init.constant_(self.final_layer.linear.bias, 0)

    def forward(self, x, t, c):
        """
        Apply the model to an input batch.
        :param x: an [N x C] Tensor of inputs.
        :param t: a 1-D batch of timesteps.
        :param c: conditioning from AR transformer.
        :return: an [N x C] Tensor of outputs.
        """
        x = self.input_proj(x)
        t = self.time_embed(t)
        c = self.cond_embed(c)

        y = t + c

        if self.grad_checkpointing and not torch.jit.is_scripting():
            for block in self.res_blocks:
                x = checkpoint(block, x, y)
        else:
            for block in self.res_blocks:
                x = block(x, y)

        return self.final_layer(x, y)

    def forward_with_cfg(self, x, t, c, cfg_scale):
        half = x[: len(x) // 2]
        combined = torch.cat([half, half], dim=0)
        model_out = self.forward(combined, t, c)
        eps, rest = model_out[:, :self.in_channels], model_out[:, self.in_channels:]
        cond_eps, uncond_eps = torch.split(eps, len(eps) // 2, dim=0)
        half_eps = uncond_eps + cfg_scale * (cond_eps - uncond_eps)
        eps = torch.cat([half_eps, half_eps], dim=0)
        return torch.cat([eps, rest], dim=1)
    
# """Vector quantizer.

# Copyright (2024) Bytedance Ltd. and/or its affiliates

# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at 

#     http://www.apache.org/licenses/LICENSE-2.0 

# Unless required by applicable law or agreed to in writing, software 
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License.

# Reference: 
#     https://github.com/CompVis/taming-transformers/blob/master/taming/modules/vqvae/quantize.py
#     https://github.com/google-research/magvit/blob/main/videogvt/models/vqvae.py
#     https://github.com/CompVis/latent-diffusion/blob/main/ldm/modules/distributions/distributions.py
#     https://github.com/lyndonzheng/CVQ-VAE/blob/main/quantise.py
# """
# from typing import Mapping, Text, Tuple

# import torch
# from einops import rearrange
# from accelerate.utils.operations import gather
# from torch.cuda.amp import autocast

class VectorQuantizer(torch.nn.Module):
    def __init__(self,
                 codebook_size: int = 1024,
                 token_size: int = 256,
                 commitment_cost: float = 0.25,
                 use_l2_norm: bool = False,
                 clustering_vq: bool = False
                 ):
        super().__init__()
        self.codebook_size = codebook_size
        self.token_size = token_size
        self.commitment_cost = commitment_cost

        self.embedding = torch.nn.Embedding(codebook_size, token_size)
        self.embedding.weight.data.uniform_(-1.0 / codebook_size, 1.0 / codebook_size)
        self.use_l2_norm = use_l2_norm

        self.clustering_vq = clustering_vq
        if clustering_vq:
            self.decay = 0.99
            self.register_buffer("embed_prob", torch.zeros(self.codebook_size))

    # Ensure quantization is performed using f32
    @autocast('cuda', enabled=False)
    def forward(self, z: torch.Tensor) -> Tuple[torch.Tensor, Mapping[Text, torch.Tensor]]:
        z = z.float()
        z = rearrange(z, 'b c h w -> b h w c').contiguous()
        z_flattened = rearrange(z, 'b h w c -> (b h w) c')
        unnormed_z_flattened = z_flattened

        if self.use_l2_norm:
            z_flattened = torch.nn.functional.normalize(z_flattened, dim=-1)
            embedding = torch.nn.functional.normalize(self.embedding.weight, dim=-1)
        else:
            embedding = self.embedding.weight
        d = torch.sum(z_flattened**2, dim=1, keepdim=True) + \
            torch.sum(embedding**2, dim=1) - 2 * \
            torch.einsum('bd,dn->bn', z_flattened, embedding.T)

        min_encoding_indices = torch.argmin(d, dim=1) # num_ele
        z_quantized = self.get_codebook_entry(min_encoding_indices).view(z.shape)

        if self.use_l2_norm:
            z = torch.nn.functional.normalize(z, dim=-1)

        # compute loss for embedding
        commitment_loss = self.commitment_cost * torch.mean((z_quantized.detach() - z) **2)
        codebook_loss = torch.mean((z_quantized - z.detach()) **2)

        if self.clustering_vq and self.training:
            with torch.no_grad():
                # Gather distance matrix from all GPUs.
                encoding_indices = gather(min_encoding_indices)
                if len(min_encoding_indices.shape) != 1:
                    raise ValueError(f"min_encoding_indices in a wrong shape, {min_encoding_indices.shape}")
                # Compute and update the usage of each entry in the codebook.
                encodings = torch.zeros(encoding_indices.shape[0], self.codebook_size, device=z.device)
                encodings.scatter_(1, encoding_indices.unsqueeze(1), 1)
                avg_probs = torch.mean(encodings, dim=0)
                self.embed_prob.mul_(self.decay).add_(avg_probs, alpha=1-self.decay)
                # Closest sampling to update the codebook.
                all_d = gather(d)
                all_unnormed_z_flattened = gather(unnormed_z_flattened).detach()
                if all_d.shape[0] != all_unnormed_z_flattened.shape[0]:
                    raise ValueError(
                        "all_d and all_unnormed_z_flattened have different length" + 
                        f"{all_d.shape}, {all_unnormed_z_flattened.shape}")
                indices = torch.argmin(all_d, dim=0)
                random_feat = all_unnormed_z_flattened[indices]
                # Decay parameter based on the average usage.
                decay = torch.exp(-(self.embed_prob * self.codebook_size * 10) /
                                   (1 - self.decay) - 1e-3).unsqueeze(1).repeat(1, self.token_size)
                self.embedding.weight.data = self.embedding.weight.data * (1 - decay) + random_feat * decay

        loss = commitment_loss + codebook_loss

        # preserve gradients
        z_quantized = z + (z_quantized - z).detach()

        # reshape back to match original input shape
        z_quantized = rearrange(z_quantized, 'b h w c -> b c h w').contiguous()

        result_dict = dict(
            quantizer_loss=loss,
            commitment_loss=commitment_loss,
            codebook_loss=codebook_loss,
            min_encoding_indices=min_encoding_indices.view(z_quantized.shape[0], z_quantized.shape[2], z_quantized.shape[3])
        )

        return z_quantized, loss, (None, None, min_encoding_indices)

    def get_codebook_entry(self, indices):
        if len(indices.shape) == 1:
            z_quantized = self.embedding(indices)
        elif len(indices.shape) == 2:
            z_quantized = torch.einsum('bd,dn->bn', indices, self.embedding.weight)
        else:
            raise NotImplementedError
        if self.use_l2_norm:
            z_quantized = torch.nn.functional.normalize(z_quantized, dim=-1)
        return z_quantized
    

# class DiagonalGaussianDistribution(object):
#     @autocast(enabled=False)
#     def __init__(self, parameters, deterministic=False):
#         """Initializes a Gaussian distribution instance given the parameters.

#         Args:
#             parameters (torch.Tensor): The parameters for the Gaussian distribution. It is expected
#                 to be in shape [B, 2 * C, *], where B is batch size, and C is the embedding dimension.
#                 First C channels are used for mean and last C are used for logvar in the Gaussian distribution.
#             deterministic (bool): Whether to use deterministic sampling. When it is true, the sampling results
#                 is purely based on mean (i.e., std = 0).
#         """
#         self.parameters = parameters
#         self.mean, self.logvar = torch.chunk(parameters.float(), 2, dim=1)
#         self.logvar = torch.clamp(self.logvar, -30.0, 20.0)
#         self.deterministic = deterministic
#         self.std = torch.exp(0.5 * self.logvar)
#         self.var = torch.exp(self.logvar)
#         if self.deterministic:
#             self.var = self.std = torch.zeros_like(self.mean).to(device=self.parameters.device)

#     @autocast(enabled=False)
#     def sample(self):
#         x = self.mean.float() + self.std.float() * torch.randn(self.mean.shape).to(device=self.parameters.device)
#         return x

#     @autocast(enabled=False)
#     def mode(self):
#         return self.mean

#     @autocast(enabled=False)
#     def kl(self):
#         if self.deterministic:
#             return torch.Tensor([0.])
#         else:
#             return 0.5 * torch.sum(torch.pow(self.mean.float(), 2)
#                                     + self.var.float() - 1.0 - self.logvar.float(),
#                                     dim=[1, 2])
        


# Conv2D with same padding
class Conv2dSame(nn.Conv2d):
    def calc_same_pad(self, i: int, k: int, s: int, d: int) -> int:
        return max((math.ceil(i / s) - 1) * s + (k - 1) * d + 1 - i, 0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        ih, iw = x.size()[-2:]

        pad_h = self.calc_same_pad(i=ih, k=self.kernel_size[0], s=self.stride[0], d=self.dilation[0])
        pad_w = self.calc_same_pad(i=iw, k=self.kernel_size[1], s=self.stride[1], d=self.dilation[1])

        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        return super().forward(x)


class ResnetBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int = None,
        dropout_prob: float = 0.0,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.out_channels_ = self.in_channels if self.out_channels is None else self.out_channels

        self.norm1 = nn.GroupNorm(num_groups=32, num_channels=in_channels, eps=1e-6, affine=True)
        self.conv1 = Conv2dSame(self.in_channels, self.out_channels_, kernel_size=3, bias=False)

        self.norm2 = nn.GroupNorm(num_groups=32, num_channels=self.out_channels_, eps=1e-6, affine=True)
        self.dropout = nn.Dropout(dropout_prob)
        self.conv2 = Conv2dSame(self.out_channels_, self.out_channels_, kernel_size=3, bias=False)

        if self.in_channels != self.out_channels_:
            self.nin_shortcut = Conv2dSame(self.out_channels_, self.out_channels_, kernel_size=1, bias=False)

    def forward(self, hidden_states):
        residual = hidden_states
        hidden_states = self.norm1(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.conv1(hidden_states)

        hidden_states = self.norm2(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.conv2(hidden_states)

        if self.in_channels != self.out_channels_:
            residual = self.nin_shortcut(hidden_states)

        return hidden_states + residual


class DownsamplingBlock(nn.Module):
    def __init__(self, config, block_idx: int):
        super().__init__()

        self.config = config
        self.block_idx = block_idx

        in_channel_mult = (1,) + tuple(self.config.channel_mult)
        block_in = self.config.hidden_channels * in_channel_mult[self.block_idx]
        block_out = self.config.hidden_channels * self.config.channel_mult[self.block_idx]

        res_blocks = nn.ModuleList()
        for _ in range(self.config.num_res_blocks):
            res_blocks.append(ResnetBlock(block_in, block_out, dropout_prob=self.config.dropout))
            block_in = block_out
        self.block = res_blocks

        self.downsample = self.block_idx != self.config.num_resolutions - 1

    def forward(self, hidden_states):
        for res_block in self.block:
            hidden_states = res_block(hidden_states)

        if self.downsample:
            hidden_states = F.avg_pool2d(hidden_states, kernel_size=2, stride=2)

        return hidden_states


class UpsamplingBlock(nn.Module):
    def __init__(self, config, block_idx: int):
        super().__init__()

        self.config = config
        self.block_idx = block_idx

        if self.block_idx == self.config.num_resolutions - 1:
            block_in = self.config.hidden_channels * self.config.channel_mult[-1]
        else:
            block_in = self.config.hidden_channels * self.config.channel_mult[self.block_idx + 1]

        block_out = self.config.hidden_channels * self.config.channel_mult[self.block_idx]

        res_blocks = []
        for _ in range(self.config.num_res_blocks):
            res_blocks.append(ResnetBlock(block_in, block_out, dropout_prob=self.config.dropout))
            block_in = block_out
        self.block = nn.ModuleList(res_blocks)

        self.add_upsample = self.block_idx != 0
        if self.add_upsample:
            self.upsample_conv = Conv2dSame(block_out, block_out, kernel_size=3)

    def forward(self, hidden_states):
        for res_block in self.block:
            hidden_states = res_block(hidden_states)

        if self.add_upsample:
            hidden_states = F.interpolate(hidden_states, scale_factor=2.0, mode="nearest")
            hidden_states = self.upsample_conv(hidden_states)

        return hidden_states

class Pixel_Encoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        # downsampling
        self.conv_in = Conv2dSame(self.config.num_channels, self.config.hidden_channels, kernel_size=3, bias=False)

        downsample_blocks = []
        for i_level in range(self.config.num_resolutions):
            downsample_blocks.append(DownsamplingBlock(self.config, block_idx=i_level))
        self.down = nn.ModuleList(downsample_blocks)

        # middle
        mid_channels = self.config.hidden_channels * self.config.channel_mult[-1]
        res_blocks = nn.ModuleList()
        for _ in range(self.config.num_res_blocks):
            res_blocks.append(ResnetBlock(mid_channels, mid_channels, dropout_prob=self.config.dropout))
        self.mid = res_blocks

        # end
        self.norm_out = nn.GroupNorm(num_groups=32, num_channels=mid_channels, eps=1e-6, affine=True)
        self.conv_out = Conv2dSame(mid_channels, self.config.z_channels, kernel_size=1)

    def forward(self, pixel_values):
        # downsampling
        hidden_states = self.conv_in(pixel_values)
        for block in self.down:
            hidden_states = block(hidden_states)

        # middle
        for block in self.mid:
            hidden_states = block(hidden_states)

        # end
        hidden_states = self.norm_out(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.conv_out(hidden_states)
        return hidden_states


class Pixel_Decoder(nn.Module):
    def __init__(self, config):
        super().__init__()

        self.config = config

        # compute in_channel_mult, block_in and curr_res at lowest res
        block_in = self.config.hidden_channels * self.config.channel_mult[self.config.num_resolutions - 1]
        curr_res = self.config.resolution // 2 ** (self.config.num_resolutions - 1)
        self.z_shape = (1, self.config.z_channels, curr_res, curr_res)

        # z to block_in
        self.conv_in = Conv2dSame(self.config.z_channels, block_in, kernel_size=3)

        # middle
        res_blocks = nn.ModuleList()
        for _ in range(self.config.num_res_blocks):
            res_blocks.append(ResnetBlock(block_in, block_in, dropout_prob=self.config.dropout))
        self.mid = res_blocks

        # upsampling
        upsample_blocks = []
        for i_level in reversed(range(self.config.num_resolutions)):
            upsample_blocks.append(UpsamplingBlock(self.config, block_idx=i_level))
        self.up = nn.ModuleList(list(reversed(upsample_blocks)))  # reverse to get consistent order

        # end
        block_out = self.config.hidden_channels * self.config.channel_mult[0]
        self.norm_out = nn.GroupNorm(num_groups=32, num_channels=block_out, eps=1e-6, affine=True)
        self.conv_out = Conv2dSame(block_out, self.config.num_channels, kernel_size=3)

    def forward(self, hidden_states):
        # z to block_in
        hidden_states = self.conv_in(hidden_states)

        # middle
        for block in self.mid:
            hidden_states = block(hidden_states)

        # upsampling
        for block in reversed(self.up):
            hidden_states = block(hidden_states)

        # end
        hidden_states = self.norm_out(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.conv_out(hidden_states)

        return hidden_states


class Pixel_Quantizer(nn.Module):
    """
    see https://github.com/MishaLaskin/vqvae/blob/d761a999e2267766400dc646d82d3ac3657771d4/models/quantizer.py
    Discretization bottleneck part of the VQ-VAE.
    """

    def __init__(self, num_embeddings, embedding_dim, commitment_cost):
        r"""
        Args:
            num_embeddings: number of vectors in the quantized space.
            embedding_dim: dimensionality of the tensors in the quantized space.
                Inputs to the modules must be in this format as well.
            commitment_cost: scalar which controls the weighting of the loss terms
                (see equation 4 in the paper https://arxiv.org/abs/1711.00937 - this variable is Beta).
        """
        super().__init__()

        self.num_embeddings = num_embeddings
        self.n_e = num_embeddings
        self.embedding_dim = embedding_dim
        self.e_dim = embedding_dim
        self.commitment_cost = commitment_cost
        self.beta = commitment_cost

        self.embedding = nn.Embedding(num_embeddings, embedding_dim)
        self.embedding.weight.data.uniform_(-1.0 / num_embeddings, 1.0 / num_embeddings)

    def forward(self, hidden_states, return_loss=False):
        """
        Inputs the output of the encoder network z and maps it to a discrete one-hot vector that is the index of the
        closest embedding vector e_j z (continuous) -> z_q (discrete) z.shape = (batch, channel, height, width)
        quantization pipeline:
            1. get encoder input (B,C,H,W)
            2. flatten input to (B*H*W,C)
        """
        # reshape z -> (batch, height, width, channel) and flatten
        hidden_states = hidden_states.permute(0, 2, 3, 1).contiguous()

        distances = self.compute_distances(hidden_states)
        min_encoding_indices = torch.argmin(distances, axis=1).unsqueeze(1)
        min_encodings = torch.zeros(min_encoding_indices.shape[0], self.num_embeddings).to(hidden_states)
        min_encodings.scatter_(1, min_encoding_indices, 1)

        # get quantized latent vectors
        z_q = torch.matmul(min_encodings, self.embedding.weight).view(hidden_states.shape)

        # reshape to (batch, num_tokens)
        min_encoding_indices = min_encoding_indices.reshape(hidden_states.shape[0], -1)

        # compute loss for embedding
        loss = None
        if return_loss:
            loss = torch.mean((z_q.detach() - hidden_states) ** 2) + self.commitment_cost * torch.mean(
                (z_q - hidden_states.detach()) ** 2
            )
            # preserve gradients
            z_q = hidden_states + (z_q - hidden_states).detach()

        # reshape back to match original input shape
        z_q = z_q.permute(0, 3, 1, 2).contiguous()

        return z_q, loss, (None,None, min_encoding_indices)  # only interface adaption

    def compute_distances(self, hidden_states):
        # distances from z to embeddings e_j (z - e)^2 = z^2 + e^2 - 2 e * z
        hidden_states_flattended = hidden_states.reshape((-1, self.embedding_dim))
        emb_weights = self.embedding.weight.t()

        inputs_norm_sq = hidden_states_flattended.pow(2.0).sum(dim=1, keepdim=True)
        codebook_t_norm_sq = emb_weights.pow(2.0).sum(dim=0, keepdim=True)
        distances = torch.addmm(
            inputs_norm_sq + codebook_t_norm_sq,
            hidden_states_flattended,
            emb_weights,
            alpha=-2.0,
        )
        return distances

    def get_codebook_entry(self, indices):
        # indices are expected to be of shape (batch, num_tokens)
        # get quantized latent vectors
        if len(indices.shape) == 2:
            batch, num_tokens = indices.shape
            z_q = self.embedding(indices)
            z_q = z_q.reshape(batch, int(math.sqrt(num_tokens)), int(math.sqrt(num_tokens)), -1).permute(0, 3, 1, 2)
        elif len(indices.shape) == 3:
            batch, height, width = indices.shape
            indices = indices.view(batch, -1)
            z_q = self.embedding(indices)
            z_q = z_q.reshape(batch, height, width, -1).permute(0, 3, 1, 2)
        else:
            raise NotImplementedError
        return z_q

    # adapted from https://github.com/kakaobrain/rq-vae-transformer/blob/main/rqvae/models/rqvae/quantizations.py#L372
    def get_soft_code(self, hidden_states, temp=1.0, stochastic=False):
        hidden_states = hidden_states.permute(0, 2, 3, 1).contiguous()  # (batch, height, width, channel)
        distances = self.compute_distances(hidden_states)  # (batch * height * width, num_embeddings)

        soft_code = F.softmax(-distances / temp, dim=-1)  # (batch * height * width, num_embeddings)
        if stochastic:
            code = torch.multinomial(soft_code, 1)  # (batch * height * width, 1)
        else:
            code = distances.argmin(dim=-1)  # (batch * height * width)

        code = code.reshape(hidden_states.shape[0], -1)  # (batch, height * width)
        batch, num_tokens = code.shape
        soft_code = soft_code.reshape(batch, num_tokens, -1)  # (batch, height * width, num_embeddings)
        return soft_code, code

    def get_code(self, hidden_states):
        # reshape z -> (batch, height, width, channel)
        hidden_states = hidden_states.permute(0, 2, 3, 1).contiguous()
        distances = self.compute_distances(hidden_states)
        indices = torch.argmin(distances, axis=1).unsqueeze(1)
        indices = indices.reshape(hidden_states.shape[0], -1)
        return indices



import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint
import math

from medlat.diffusion import create_gaussian_diffusion


class DiffLoss(nn.Module):
    """Diffusion Loss"""
    def __init__(self, target_channels, z_channels, depth, width, num_sampling_steps, grad_checkpointing=False, diffusion_batch_mul=4, noise_schedule="cosine", learn_sigma=True):
        super(DiffLoss, self).__init__()
        self.in_channels = target_channels
        self.net = SimpleMLPAdaLN(
            in_channels=target_channels,
            model_channels=width,
            out_channels=target_channels * 2,  # for vlb loss
            z_channels=z_channels,
            num_res_blocks=depth,
            grad_checkpointing=grad_checkpointing
        )
        self.diffusion_batch_mul = diffusion_batch_mul
        self.train_diffusion = create_gaussian_diffusion(timestep_respacing="", noise_schedule=noise_schedule, learn_sigma=learn_sigma)
        self.gen_diffusion = create_gaussian_diffusion(timestep_respacing=num_sampling_steps, noise_schedule=noise_schedule, learn_sigma=learn_sigma)

    def forward(self, target, z, mask=None):
        ## diffusion batch multiplier to make better use of memory and compute
        bsz, seq_len, _ = target.shape
        target = target.reshape(bsz * seq_len, -1).repeat(self.diffusion_batch_mul, 1)
        z = z.reshape(bsz*seq_len, -1).repeat(self.diffusion_batch_mul, 1)
        if mask is not None:
            mask = mask.reshape(bsz*seq_len).repeat(self.diffusion_batch_mul)

        t = torch.randint(0, self.train_diffusion.num_timesteps, (target.shape[0],), device=target.device)
        model_kwargs = dict(c=z)
        loss_dict = self.train_diffusion.training_losses(self.net, target, t, model_kwargs)
        loss = loss_dict["loss"]
        if mask is not None:
            loss = (loss * mask).sum() / mask.sum()
        return loss.mean()

    def sample(self, z, temperature=1.0, cfg=1.0):
        # diffusion loss sampling
        
        if not cfg == 1.0:
            noise = torch.randn(z.shape[0] // 2, self.in_channels).cuda()
            noise = torch.cat([noise, noise], dim=0)
            model_kwargs = dict(c=z, cfg_scale=cfg)
            sample_fn = self.net.forward_with_cfg
        else:
            noise = torch.randn(z.shape[0], self.in_channels).cuda()
            model_kwargs = dict(c=z)
            sample_fn = self.net.forward

        sampled_token_latent = self.gen_diffusion.p_sample_loop(
            sample_fn, noise.shape, noise, clip_denoised=False, model_kwargs=model_kwargs, progress=False, device=z.device,
            temperature=temperature
        )

        return sampled_token_latent


def modulate(x, shift, scale):
    return x * (1 + scale) + shift


class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """
    def __init__(self, hidden_size, frequency_embedding_size=256):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """
        # https://github.com/openai/glide-text2im/blob/main/glide_text2im/nn.py
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=t.device)
        args = t[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq)
        return t_emb


class ResBlock(nn.Module):
    """
    A residual block that can optionally change the number of channels.
    :param channels: the number of input channels.
    """

    def __init__(
        self,
        channels
    ):
        super().__init__()
        self.channels = channels

        self.in_ln = nn.LayerNorm(channels, eps=1e-6)
        self.mlp = nn.Sequential(
            nn.Linear(channels, channels, bias=True),
            nn.SiLU(),
            nn.Linear(channels, channels, bias=True),
        )

        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(channels, 3 * channels, bias=True)
        )

    def forward(self, x, y):
        shift_mlp, scale_mlp, gate_mlp = self.adaLN_modulation(y).chunk(3, dim=-1)
        h = modulate(self.in_ln(x), shift_mlp, scale_mlp)
        h = self.mlp(h)
        return x + gate_mlp * h


class FinalLayer(nn.Module):
    """
    The final layer of DiT.
    """
    def __init__(self, model_channels, out_channels):
        super().__init__()
        self.norm_final = nn.LayerNorm(model_channels, elementwise_affine=False, eps=1e-6)
        self.linear = nn.Linear(model_channels, out_channels, bias=True)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(model_channels, 2 * model_channels, bias=True)
        )

    def forward(self, x, c):
        shift, scale = self.adaLN_modulation(c).chunk(2, dim=-1)
        x = modulate(self.norm_final(x), shift, scale)
        x = self.linear(x)
        return x


class SimpleMLPAdaLN(nn.Module):
    """
    The MLP for Diffusion Loss.
    :param in_channels: channels in the input Tensor.
    :param model_channels: base channel count for the model.
    :param out_channels: channels in the output Tensor.
    :param z_channels: channels in the condition.
    :param num_res_blocks: number of residual blocks per downsample.
    """

    def __init__(
        self,
        in_channels,
        model_channels,
        out_channels,
        z_channels,
        num_res_blocks,
        grad_checkpointing=False
    ):
        super().__init__()

        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        self.num_res_blocks = num_res_blocks
        self.grad_checkpointing = grad_checkpointing

        self.time_embed = TimestepEmbedder(model_channels)
        self.cond_embed = nn.Linear(z_channels, model_channels)

        self.input_proj = nn.Linear(in_channels, model_channels)

        res_blocks = []
        for i in range(num_res_blocks):
            res_blocks.append(ResBlock(
                model_channels,
            ))

        self.res_blocks = nn.ModuleList(res_blocks)
        self.final_layer = FinalLayer(model_channels, out_channels)

        self.initialize_weights()

    def initialize_weights(self):
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)
        self.apply(_basic_init)

        # Initialize timestep embedding MLP
        nn.init.normal_(self.time_embed.mlp[0].weight, std=0.02)
        nn.init.normal_(self.time_embed.mlp[2].weight, std=0.02)

        # Zero-out adaLN modulation layers
        for block in self.res_blocks:
            nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(block.adaLN_modulation[-1].bias, 0)

        # Zero-out output layers
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].bias, 0)
        nn.init.constant_(self.final_layer.linear.weight, 0)
        nn.init.constant_(self.final_layer.linear.bias, 0)

    def forward(self, x, t, c):
        """
        Apply the model to an input batch.
        :param x: an [N x C x ...] Tensor of inputs.
        :param t: a 1-D batch of timesteps.
        :param c: conditioning from AR transformer.
        :return: an [N x C x ...] Tensor of outputs.
        """
        # print(c.shape)
        x = self.input_proj(x)
        t = self.time_embed(t)
        c = self.cond_embed(c)

        y = t + c

        if self.grad_checkpointing and not torch.jit.is_scripting():
            for block in self.res_blocks:
                x = checkpoint(block, x, y, use_reentrant=False)
        else:
            for block in self.res_blocks:
                x = block(x, y)

        return self.final_layer(x, y)

    def forward_with_cfg(self, x, t, c, cfg_scale):
        half = x[: len(x) // 2]
        combined = torch.cat([half, half], dim=0)
        model_out = self.forward(combined, t, c)
        eps, rest = model_out[:, :self.in_channels], model_out[:, self.in_channels:]
        cond_eps, uncond_eps = torch.split(eps, len(eps) // 2, dim=0)
        half_eps = uncond_eps + cfg_scale * (cond_eps - uncond_eps)
        eps = torch.cat([half_eps, half_eps], dim=0)
        return torch.cat([eps, rest], dim=1)


from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .act import build_act
from .norm import build_norm
from ..utils import get_same_padding, list_sum, resize, val2list, val2tuple


def _spatial_unshuffle_nd(x: torch.Tensor, factor: int, dims: int) -> torch.Tensor:
    """Spatial unshuffle: (B, C, ...) -> (B, C * factor^spatial_dims, .../factor). Works for 2D and 3D."""
    if dims == 2:
        return F.pixel_unshuffle(x, factor)
    # 3D: (B, C, D, H, W) -> (B, C * factor^3, D//f, H//f, W//f)
    B, C, D, H, W = x.shape
    x = x.view(B, C, D // factor, factor, H // factor, factor, W // factor, factor)
    x = x.permute(0, 1, 3, 5, 7, 2, 4, 6).contiguous()
    return x.view(B, C * (factor ** 3), D // factor, H // factor, W // factor)


def _spatial_shuffle_nd(x: torch.Tensor, factor: int, dims: int) -> torch.Tensor:
    """Spatial shuffle (inverse of unshuffle): (B, C, ...) -> (B, C // factor^spatial_dims, ...*factor). Works for 2D and 3D."""
    if dims == 2:
        return F.pixel_shuffle(x, factor)
    # 3D: (B, C, D, H, W) -> (B, C // factor^3, D*f, H*f, W*f)
    B, C, D, H, W = x.shape
    x = x.view(B, C // (factor ** 3), factor, factor, factor, D, H, W)
    x = x.permute(0, 1, 5, 2, 6, 3, 7, 4).contiguous()  # (B, C', f, f, f, D, H, W) -> (B, C', D, f, H, f, W, f)
    return x.view(B, C // (factor ** 3), D * factor, H * factor, W * factor)


__all__ = [
    "ConvLayer",
    "UpSampleLayer",
    "ConvPixelUnshuffleDownSampleLayer",
    "PixelUnshuffleChannelAveragingDownSampleLayer",
    "ConvPixelShuffleUpSampleLayer",
    "ChannelDuplicatingPixelUnshuffleUpSampleLayer",
    "LinearLayer",
    "IdentityLayer",
    "DSConv",
    "MBConv",
    "FusedMBConv",
    "ResBlock",
    "LiteMLA",
    "EfficientViTBlock",
    "ResidualBlock",
    "DAGBlock",
    "OpSequential",
]


#################################################################################
#                             Basic Layers                                      #
#################################################################################


class ConvLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        dilation=1,
        groups=1,
        use_bias=False,
        dropout=0,
        norm="bn2d",
        act_func="relu",
        dims=2,
    ):
        super(ConvLayer, self).__init__()
        self.dims = dims

        pad_val = get_same_padding(kernel_size)
        pad_tuple = (pad_val,) * dims if not isinstance(pad_val, tuple) else pad_val
        padding = tuple(p * dilation for p in pad_tuple)

        self.dropout = (nn.Dropout3d(dropout, inplace=False) if dropout > 0 else None) if dims == 3 else (nn.Dropout2d(dropout, inplace=False) if dropout > 0 else None)
        if dims == 2:
            self.conv = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=(kernel_size, kernel_size),
                stride=(stride, stride),
                padding=padding,
                dilation=(dilation, dilation),
                groups=groups,
                bias=use_bias,
            )
        else:
            self.conv = nn.Conv3d(
                in_channels,
                out_channels,
                kernel_size=(kernel_size,) * 3,
                stride=(stride,) * 3,
                padding=padding,
                dilation=(dilation,) * 3,
                groups=groups,
                bias=use_bias,
            )
        self.norm = build_norm(norm, num_features=out_channels, dims=dims)
        self.act = build_act(act_func)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.dropout is not None:
            x = self.dropout(x)
        x = self.conv(x)
        if self.norm:
            x = self.norm(x)
        if self.act:
            x = self.act(x)
        return x


class UpSampleLayer(nn.Module):
    def __init__(
        self,
        mode="bicubic",
        size: Optional[int | tuple[int, int] | list[int]] = None,
        factor=2,
        align_corners=False,
    ):
        super(UpSampleLayer, self).__init__()
        self.mode = mode
        self.size = val2list(size, 2) if size is not None else None
        self.factor = None if self.size is not None else factor
        self.align_corners = align_corners

    @torch.autocast(device_type="cuda", enabled=False)
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if (self.size is not None and tuple(x.shape[-2:]) == self.size) or self.factor == 1:
            return x
        if x.dtype in [torch.float16, torch.bfloat16]:
            x = x.float()
        return resize(x, self.size, self.factor, self.mode, self.align_corners)


class ConvPixelUnshuffleDownSampleLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        factor: int,
        dims: int = 2,
    ):
        super().__init__()
        self.factor = factor
        self.dims = dims
        out_ratio = factor ** (2 if dims == 2 else 3)
        assert out_channels % out_ratio == 0
        self.conv = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels // out_ratio,
            kernel_size=kernel_size,
            use_bias=True,
            norm=None,
            act_func=None,
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = _spatial_unshuffle_nd(x, self.factor, self.dims)
        return x


class PixelUnshuffleChannelAveragingDownSampleLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        factor: int,
        dims: int = 2,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.factor = factor
        self.dims = dims
        spatial_pow = 2 if dims == 2 else 3
        assert in_channels * (factor ** spatial_pow) % out_channels == 0
        self.group_size = in_channels * (factor ** spatial_pow) // out_channels

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = _spatial_unshuffle_nd(x, self.factor, self.dims)
        B, C = x.shape[:2]
        spatial = x.shape[2:]
        x = x.view(B, self.out_channels, self.group_size, *spatial)
        x = x.mean(dim=2)
        return x


class ConvPixelShuffleUpSampleLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        factor: int,
        dims: int = 2,
    ):
        super().__init__()
        self.factor = factor
        self.dims = dims
        out_ratio = factor ** (2 if dims == 2 else 3)
        self.conv = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels * out_ratio,
            kernel_size=kernel_size,
            use_bias=True,
            norm=None,
            act_func=None,
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = _spatial_shuffle_nd(x, self.factor, self.dims)
        return x


class InterpolateConvUpSampleLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        factor: int,
        mode: str = "nearest",
        dims: int = 2,
    ) -> None:
        super().__init__()
        self.factor = factor
        self.mode = mode
        self.dims = dims
        self.conv = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            use_bias=True,
            norm=None,
            act_func=None,
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.dims == 3:
            scale = (self.factor,) * 3
            x = torch.nn.functional.interpolate(x, scale_factor=scale, mode="trilinear" if self.mode == "bilinear" else self.mode)
        else:
            x = torch.nn.functional.interpolate(x, scale_factor=self.factor, mode=self.mode)
        x = self.conv(x)
        return x


class ChannelDuplicatingPixelUnshuffleUpSampleLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        factor: int,
        dims: int = 2,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.factor = factor
        self.dims = dims
        spatial_pow = 2 if dims == 2 else 3
        assert out_channels * (factor ** spatial_pow) % in_channels == 0
        self.repeats = out_channels * (factor ** spatial_pow) // in_channels

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat_interleave(self.repeats, dim=1)
        x = _spatial_shuffle_nd(x, self.factor, self.dims)
        return x


class LinearLayer(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        use_bias=True,
        dropout=0,
        norm=None,
        act_func=None,
    ):
        super(LinearLayer, self).__init__()

        self.dropout = nn.Dropout(dropout, inplace=False) if dropout > 0 else None
        self.linear = nn.Linear(in_features, out_features, use_bias)
        self.norm = build_norm(norm, num_features=out_features)
        self.act = build_act(act_func)

    def _try_squeeze(self, x: torch.Tensor) -> torch.Tensor:
        if x.dim() > 2:
            x = torch.flatten(x, start_dim=1)
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._try_squeeze(x)
        if self.dropout:
            x = self.dropout(x)
        x = self.linear(x)
        if self.norm:
            x = self.norm(x)
        if self.act:
            x = self.act(x)
        return x


class IdentityLayer(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


#################################################################################
#                             Basic Blocks                                      #
#################################################################################


class DSConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        use_bias=False,
        norm=("bn2d", "bn2d"),
        act_func=("relu6", None),
        dims=2,
    ):
        super(DSConv, self).__init__()

        use_bias = val2tuple(use_bias, 2)
        norm = val2tuple(norm, 2)
        act_func = val2tuple(act_func, 2)

        self.depth_conv = ConvLayer(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            groups=in_channels,
            norm=norm[0],
            act_func=act_func[0],
            use_bias=use_bias[0],
            dims=dims,
        )
        self.point_conv = ConvLayer(
            in_channels,
            out_channels,
            1,
            norm=norm[1],
            act_func=act_func[1],
            use_bias=use_bias[1],
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.depth_conv(x)
        x = self.point_conv(x)
        return x


class MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        mid_channels=None,
        expand_ratio=6,
        use_bias=False,
        norm=("bn2d", "bn2d", "bn2d"),
        act_func=("relu6", "relu6", None),
        dims=2,
    ):
        super(MBConv, self).__init__()

        use_bias = val2tuple(use_bias, 3)
        norm = val2tuple(norm, 3)
        act_func = val2tuple(act_func, 3)
        mid_channels = round(in_channels * expand_ratio) if mid_channels is None else mid_channels

        self.inverted_conv = ConvLayer(
            in_channels,
            mid_channels,
            1,
            stride=1,
            norm=norm[0],
            act_func=act_func[0],
            use_bias=use_bias[0],
            dims=dims,
        )
        self.depth_conv = ConvLayer(
            mid_channels,
            mid_channels,
            kernel_size,
            stride=stride,
            groups=mid_channels,
            norm=norm[1],
            act_func=act_func[1],
            use_bias=use_bias[1],
            dims=dims,
        )
        self.point_conv = ConvLayer(
            mid_channels,
            out_channels,
            1,
            norm=norm[2],
            act_func=act_func[2],
            use_bias=use_bias[2],
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.inverted_conv(x)
        x = self.depth_conv(x)
        x = self.point_conv(x)
        return x


class FusedMBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        mid_channels=None,
        expand_ratio=6,
        groups=1,
        use_bias=False,
        norm=("bn2d", "bn2d"),
        act_func=("relu6", None),
        dims=2,
    ):
        super().__init__()
        use_bias = val2tuple(use_bias, 2)
        norm = val2tuple(norm, 2)
        act_func = val2tuple(act_func, 2)

        mid_channels = round(in_channels * expand_ratio) if mid_channels is None else mid_channels

        self.spatial_conv = ConvLayer(
            in_channels,
            mid_channels,
            kernel_size,
            stride,
            groups=groups,
            use_bias=use_bias[0],
            norm=norm[0],
            act_func=act_func[0],
            dims=dims,
        )
        self.point_conv = ConvLayer(
            mid_channels,
            out_channels,
            1,
            use_bias=use_bias[1],
            norm=norm[1],
            act_func=act_func[1],
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.spatial_conv(x)
        x = self.point_conv(x)
        return x


class GLUMBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        mid_channels=None,
        expand_ratio=6,
        use_bias=False,
        norm=(None, None, "ln2d"),
        act_func=("silu", "silu", None),
        dims=2,
    ):
        super().__init__()
        use_bias = val2tuple(use_bias, 3)
        norm = val2tuple(norm, 3)
        act_func = val2tuple(act_func, 3)

        mid_channels = round(in_channels * expand_ratio) if mid_channels is None else mid_channels

        self.glu_act = build_act(act_func[1], inplace=False)
        self.inverted_conv = ConvLayer(
            in_channels,
            mid_channels * 2,
            1,
            use_bias=use_bias[0],
            norm=norm[0],
            act_func=act_func[0],
            dims=dims,
        )
        self.depth_conv = ConvLayer(
            mid_channels * 2,
            mid_channels * 2,
            kernel_size,
            stride=stride,
            groups=mid_channels * 2,
            use_bias=use_bias[1],
            norm=norm[1],
            act_func=None,
            dims=dims,
        )
        self.point_conv = ConvLayer(
            mid_channels,
            out_channels,
            1,
            use_bias=use_bias[2],
            norm=norm[2],
            act_func=act_func[2],
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.inverted_conv(x)
        x = self.depth_conv(x)

        x, gate = torch.chunk(x, 2, dim=1)
        gate = self.glu_act(gate)
        x = x * gate

        x = self.point_conv(x)
        return x


class ResBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size=3,
        stride=1,
        mid_channels=None,
        expand_ratio=1,
        use_bias=False,
        norm=("bn2d", "bn2d"),
        act_func=("relu6", None),
        dims=2,
    ):
        super().__init__()
        use_bias = val2tuple(use_bias, 2)
        norm = val2tuple(norm, 2)
        act_func = val2tuple(act_func, 2)

        mid_channels = round(in_channels * expand_ratio) if mid_channels is None else mid_channels

        self.conv1 = ConvLayer(
            in_channels,
            mid_channels,
            kernel_size,
            stride,
            use_bias=use_bias[0],
            norm=norm[0],
            act_func=act_func[0],
            dims=dims,
        )
        self.conv2 = ConvLayer(
            mid_channels,
            out_channels,
            kernel_size,
            1,
            use_bias=use_bias[1],
            norm=norm[1],
            act_func=act_func[1],
            dims=dims,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class LiteMLA(nn.Module):
    r"""Lightweight multi-scale linear attention (EfficientViT, MIT Han Lab).

    This is *not* standard ViT self-attention. It produces Q,K,V via 1x1 convs,
    then applies linear/kernel attention (ReLU feature map + matmul over spatial dim)
    for O(n) context mixing. Multi-scale branches use depthwise convs. The only
    non-conv part is the matmul over spatial positions (H*W) in relu_linear_att /
    relu_quadratic_att.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        heads: Optional[int] = None,
        heads_ratio: float = 1.0,
        dim=8,
        use_bias=False,
        norm=(None, "bn2d"),
        act_func=(None, None),
        kernel_func="relu",
        scales: tuple[int, ...] = (5,),
        eps=1.0e-15,
    ):
        super(LiteMLA, self).__init__()
        self.eps = eps
        heads = int(in_channels // dim * heads_ratio) if heads is None else heads

        total_dim = heads * dim

        use_bias = val2tuple(use_bias, 2)
        norm = val2tuple(norm, 2)
        act_func = val2tuple(act_func, 2)

        self.dim = dim
        self.qkv = ConvLayer(
            in_channels,
            3 * total_dim,
            1,
            use_bias=use_bias[0],
            norm=norm[0],
            act_func=act_func[0],
        )
        self.aggreg = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Conv2d(
                        3 * total_dim,
                        3 * total_dim,
                        scale,
                        padding=get_same_padding(scale),
                        groups=3 * total_dim,
                        bias=use_bias[0],
                    ),
                    nn.Conv2d(3 * total_dim, 3 * total_dim, 1, groups=3 * heads, bias=use_bias[0]),
                )
                for scale in scales
            ]
        )
        self.kernel_func = build_act(kernel_func, inplace=False)

        self.proj = ConvLayer(
            total_dim * (1 + len(scales)),
            out_channels,
            1,
            use_bias=use_bias[1],
            norm=norm[1],
            act_func=act_func[1],
        )

    @torch.autocast(device_type="cuda", enabled=False)
    def relu_linear_att(self, qkv: torch.Tensor) -> torch.Tensor:
        B, _, H, W = list(qkv.size())

        if qkv.dtype == torch.float16:
            qkv = qkv.float()

        qkv = torch.reshape(
            qkv,
            (
                B,
                -1,
                3 * self.dim,
                H * W,
            ),
        )
        q, k, v = (
            qkv[:, :, 0 : self.dim],
            qkv[:, :, self.dim : 2 * self.dim],
            qkv[:, :, 2 * self.dim :],
        )

        # lightweight linear attention
        q = self.kernel_func(q)
        k = self.kernel_func(k)

        # linear matmul
        trans_k = k.transpose(-1, -2)

        v = F.pad(v, (0, 0, 0, 1), mode="constant", value=1)
        vk = torch.matmul(v, trans_k)
        out = torch.matmul(vk, q)
        if out.dtype == torch.bfloat16:
            out = out.float()
        out = out[:, :, :-1] / (out[:, :, -1:] + self.eps)

        out = torch.reshape(out, (B, -1, H, W))
        return out

    @torch.autocast(device_type="cuda", enabled=False)
    def relu_quadratic_att(self, qkv: torch.Tensor) -> torch.Tensor:
        B, _, H, W = list(qkv.size())

        qkv = torch.reshape(
            qkv,
            (
                B,
                -1,
                3 * self.dim,
                H * W,
            ),
        )
        q, k, v = (
            qkv[:, :, 0 : self.dim],
            qkv[:, :, self.dim : 2 * self.dim],
            qkv[:, :, 2 * self.dim :],
        )

        q = self.kernel_func(q)
        k = self.kernel_func(k)

        att_map = torch.matmul(k.transpose(-1, -2), q)  # b h n n
        original_dtype = att_map.dtype
        if original_dtype in [torch.float16, torch.bfloat16]:
            att_map = att_map.float()
        att_map = att_map / (torch.sum(att_map, dim=2, keepdim=True) + self.eps)  # b h n n
        att_map = att_map.to(original_dtype)
        out = torch.matmul(v, att_map)  # b h d n

        out = torch.reshape(out, (B, -1, H, W))
        return out

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # generate multi-scale q, k, v
        qkv = self.qkv(x)
        multi_scale_qkv = [qkv]
        for op in self.aggreg:
            multi_scale_qkv.append(op(qkv))
        qkv = torch.cat(multi_scale_qkv, dim=1)

        H, W = list(qkv.size())[-2:]
        if H * W > self.dim:
            out = self.relu_linear_att(qkv).to(qkv.dtype)
        else:
            out = self.relu_quadratic_att(qkv)
        out = self.proj(out)

        return out


class EfficientViTBlock(nn.Module):
    """EfficientViT block: hybrid of lightweight linear attention + convolutions.

    Not a standard Vision Transformer. Each block has two branches (both with residual):
    1. context_module: LiteMLA — Q,K,V from 1x1 convs, then linear attention over
       spatial positions (the only non-conv "attention" part).
    2. local_module: MBConv or GLUMBConv — inverted bottleneck + depthwise conv
       (optionally with GLU gating). Purely convolutional.

    So most of the block is convolutions; the "ViT" name comes from the EfficientViT
    family (e.g. EfficientViT: Memory Efficient Vision Transformer with Scale-wise
    Attention) which uses this lightweight attention for context mixing.
    """

    def __init__(
        self,
        in_channels: int,
        heads_ratio: float = 1.0,
        dim=32,
        expand_ratio: float = 4,
        scales: tuple[int, ...] = (5,),
        norm: str = "bn2d",
        act_func: str = "hswish",
        context_module: str = "LiteMLA",
        local_module: str = "MBConv",
    ):
        super(EfficientViTBlock, self).__init__()
        if context_module == "LiteMLA":
            self.context_module = ResidualBlock(
                LiteMLA(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    heads_ratio=heads_ratio,
                    dim=dim,
                    norm=(None, norm),
                    scales=scales,
                ),
                IdentityLayer(),
            )
        else:
            raise ValueError(f"context_module {context_module} is not supported")
        if local_module == "MBConv":
            self.local_module = ResidualBlock(
                MBConv(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    expand_ratio=expand_ratio,
                    use_bias=(True, True, False),
                    norm=(None, None, norm),
                    act_func=(act_func, act_func, None),
                ),
                IdentityLayer(),
            )
        elif local_module == "GLUMBConv":
            self.local_module = ResidualBlock(
                GLUMBConv(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    expand_ratio=expand_ratio,
                    use_bias=(True, True, False),
                    norm=(None, None, norm),
                    act_func=(act_func, act_func, None),
                ),
                IdentityLayer(),
            )
        else:
            raise NotImplementedError(f"local_module {local_module} is not supported")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.context_module(x)
        x = self.local_module(x)
        return x


#################################################################################
#                             Functional Blocks                                 #
#################################################################################


class ResidualBlock(nn.Module):
    def __init__(
        self,
        main: Optional[nn.Module],
        shortcut: Optional[nn.Module],
        post_act=None,
        pre_norm: Optional[nn.Module] = None,
    ):
        super(ResidualBlock, self).__init__()

        self.pre_norm = pre_norm
        self.main = main
        self.shortcut = shortcut
        self.post_act = build_act(post_act)

    def forward_main(self, x: torch.Tensor) -> torch.Tensor:
        if self.pre_norm is None:
            return self.main(x)
        else:
            return self.main(self.pre_norm(x))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.main is None:
            res = x
        elif self.shortcut is None:
            res = self.forward_main(x)
        else:
            res = self.forward_main(x) + self.shortcut(x)
            if self.post_act:
                res = self.post_act(res)
        return res


class DAGBlock(nn.Module):
    def __init__(
        self,
        inputs: dict[str, nn.Module],
        merge: str,
        post_input: Optional[nn.Module],
        middle: nn.Module,
        outputs: dict[str, nn.Module],
    ):
        super(DAGBlock, self).__init__()

        self.input_keys = list(inputs.keys())
        self.input_ops = nn.ModuleList(list(inputs.values()))
        self.merge = merge
        self.post_input = post_input

        self.middle = middle

        self.output_keys = list(outputs.keys())
        self.output_ops = nn.ModuleList(list(outputs.values()))

    def forward(self, feature_dict: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        feat = [op(feature_dict[key]) for key, op in zip(self.input_keys, self.input_ops)]
        if self.merge == "add":
            feat = list_sum(feat)
        elif self.merge == "cat":
            feat = torch.concat(feat, dim=1)
        else:
            raise NotImplementedError
        if self.post_input is not None:
            feat = self.post_input(feat)
        feat = self.middle(feat)
        for key, op in zip(self.output_keys, self.output_ops):
            feature_dict[key] = op(feat)
        return feature_dict


class OpSequential(nn.Module):
    def __init__(self, op_list: list[Optional[nn.Module]]):
        super(OpSequential, self).__init__()
        valid_op_list = []
        for op in op_list:
            if op is not None:
                valid_op_list.append(op)
        self.op_list = nn.ModuleList(valid_op_list)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for op in self.op_list:
            x = op(x)
        return x



from typing import Any, Optional

import numpy as np
import torch

__all__ = [
    "torch_randint",
    "torch_random",
    "torch_shuffle",
    "torch_uniform",
    "torch_random_choices",
]


def torch_randint(low: int, high: int, generator: Optional[torch.Generator] = None) -> int:
    """uniform: [low, high)"""
    if low == high:
        return low
    else:
        assert low < high
        return int(torch.randint(low=low, high=high, generator=generator, size=(1,)))


def torch_random(generator: Optional[torch.Generator] = None) -> float:
    """uniform distribution on the interval [0, 1)"""
    return float(torch.rand(1, generator=generator))


def torch_shuffle(src_list: list[Any], generator: Optional[torch.Generator] = None) -> list[Any]:
    rand_indexes = torch.randperm(len(src_list), generator=generator).tolist()
    return [src_list[i] for i in rand_indexes]


def torch_uniform(low: float, high: float, generator: Optional[torch.Generator] = None) -> float:
    """uniform distribution on the interval [low, high)"""
    rand_val = torch_random(generator)
    return (high - low) * rand_val + low


def torch_random_choices(
    src_list: list[Any],
    generator: Optional[torch.Generator] = None,
    k=1,
    weight_list: Optional[list[float]] = None,
) -> Any | list:
    if weight_list is None:
        rand_idx = torch.randint(low=0, high=len(src_list), generator=generator, size=(k,))
        out_list = [src_list[i] for i in rand_idx]
    else:
        assert len(weight_list) == len(src_list)
        accumulate_weight_list = np.cumsum(weight_list)

        out_list = []
        for _ in range(k):
            val = torch_uniform(0, accumulate_weight_list[-1], generator)
            active_id = 0
            for i, weight_val in enumerate(accumulate_weight_list):
                active_id = i
                if weight_val > val:
                    break
            out_list.append(src_list[active_id])

    return out_list[0] if k == 1 else out_list



from medlat.registry import register_model
from .models import MaskedAutoencoderViT


@register_model("token.vmae.t_p8_d16")
def VMAE_T_P8_D16(**kwargs):
    """Tiny VMAE, patch size 8, latent dim 16 (enc/dec 96d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=96,
        depth=12,
        num_heads=8,
        decoder_embed_dim=96,
        decoder_depth=12,
        decoder_num_heads=8,
        mlp_ratio=4,
        latent_dim=16,
        **kwargs,
    )


@register_model("token.vmae.t_p8_d16_asym")
def VMAE_T_P8_D16_ASYM(**kwargs):
    """Tiny-asymmetric VMAE, patch size 8, latent dim 16 (enc 96d, dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=96,
        depth=12,
        num_heads=8,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=16,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d32")
def VMAE_S_P8_D32(**kwargs):
    """Small VMAE, patch size 8, latent dim 32 (enc/dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=32,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d32_flex")
def VMAE_S_P8_D32_FLEX(**kwargs):
    """Flexible small VMAE, patch size 8, latent dim 32 (enc/dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=32,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d16")
def VMAE_S_P8_D16(**kwargs):
    """Small VMAE, patch size 8, latent dim 16 (enc 192d, dec 384d, MLP down)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=384,
        decoder_depth=12,
        decoder_num_heads=24,
        mlp_ratio=4,
        latent_dim=16,
        down_nonlinear=True,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d16_prev")
def VMAE_S_P8_D16_PREV(**kwargs):
    """Previous small VMAE variant, patch size 8, latent dim 16 (enc/dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=16,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d16_flex")
def VMAE_S_P8_D16_FLEX(**kwargs):
    """Flexible small VMAE, patch size 8, latent dim 16 (enc 192d, dec 384d, MLP down)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=384,
        decoder_depth=12,
        decoder_num_heads=24,
        mlp_ratio=4,
        latent_dim=16,
        down_nonlinear=True,
        **kwargs,
    )


@register_model("token.vmae.s_p8_d16_alt")
def VMAE_S_P8_D16_ALT(**kwargs):
    """Alternative small VMAE config, patch size 8, latent dim 16 (enc/dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=16,
        **kwargs,
    )


@register_model("token.vmae.b_p8_d16_prev")
def VMAE_B_P8_D16_PREV(**kwargs):
    """Base VMAE (previous large variant), patch size 8, latent dim 16 (enc/dec 384d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=8,
        embed_dim=384,
        depth=12,
        num_heads=16,
        decoder_embed_dim=384,
        decoder_depth=12,
        decoder_num_heads=16,
        mlp_ratio=4,
        latent_dim=16,
        **kwargs,
    )


@register_model("token.vmae.s_p16_d32")
def VMAE_S_P16_D32(**kwargs):
    """Small VMAE, patch size 16, latent dim 32 (enc/dec 192d)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=16,
        embed_dim=192,
        depth=12,
        num_heads=12,
        decoder_embed_dim=192,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=32,
        **kwargs,
    )


@register_model("token.vmae.b_p16_d32")
def VMAE_B_P16_D32(**kwargs):
    """Base VMAE, patch size 16, latent dim 32 (enc/dec 384d, downsizing at layer 4)."""
    return MaskedAutoencoderViT(
        img_size=128,
        patch_size=16,
        embed_dim=384,
        depth=12,
        num_heads=12,
        decoder_embed_dim=384,
        decoder_depth=12,
        decoder_num_heads=12,
        mlp_ratio=4,
        latent_dim=32,
        finetune_downsample_layer=4,
        **kwargs,
    )



import torch
from medlat.registry import register_model
from medlat.utils import instantiate_from_config
from .unet import UNetModel as ADM
from .unet import EncoderUNetModel as ADM_classifier

"""
Optimized configs for ADM diffusion and classifier models.
Models are now registered with method-based configurations.
"""

# ---- Diffusion Model Config Methods ----

@register_model("adm.diffusion.64C")
def adm_diffusion_64_conditioned_cfg(**overrides):
    cfg = dict(
        image_size=64,
        in_channels=3,
        model_channels=192,
        num_classes=1000,
        out_channels=6,
        num_res_blocks=3,
        attention_resolutions=[32, 16, 8],
        dropout=0.1,
        channel_mult=(1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=True,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=True,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.64U")
def adm_diffusion_64_unconditioned_cfg(**overrides):
    cfg = dict(
        image_size=64,
        in_channels=3,
        model_channels=192,
        num_classes=None,
        out_channels=6,
        num_res_blocks=3,
        attention_resolutions=[32, 16, 8],
        dropout=0.1,
        channel_mult=(1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,  # Can override as needed
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=True,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.128C")
def adm_diffusion_128_conditioned_cfg(**overrides):
    cfg = dict(
        image_size=128,
        in_channels=3,
        model_channels=256,
        num_classes=1000,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=4,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.128U")
def adm_diffusion_128_unconditioned_cfg(**overrides):
    cfg = dict(
        image_size=128,
        in_channels=3,
        model_channels=256,
        num_classes=None,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=4,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.256C")
def adm_diffusion_256_conditioned_cfg(**overrides):
    cfg = dict(
        image_size=256,
        in_channels=3,
        model_channels=256,
        num_classes=1000,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.256U")
def adm_diffusion_256_unconditioned_cfg(**overrides):
    cfg = dict(
        image_size=256,
        in_channels=3,
        model_channels=256,
        num_classes=None,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.512C")
def adm_diffusion_512_conditioned_cfg(**overrides):
    cfg = dict(
        image_size=512,
        in_channels=3,
        model_channels=256,
        num_classes=1000,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(0.5, 1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.diffusion.512U")
def adm_diffusion_512_unconditioned_cfg(**overrides):
    cfg = dict(
        image_size=512,
        in_channels=3,
        model_channels=256,
        num_classes=None,
        out_channels=6,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(0.5, 1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
    )
    cfg.update(overrides)
    return cfg

# ---- Classifier Model Config Methods ----

@register_model("adm.classifier.64C")
def adm_classifier_64_cfg(**overrides):
    cfg = dict(
        image_size=64,
        in_channels=3,
        model_channels=192,
        out_channels=1000,
        num_res_blocks=3,
        attention_resolutions=[32, 16, 8],
        dropout=0.1,
        channel_mult=(1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=True,
        pool="adaptive",
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.classifier.128C")
def adm_classifier_128_cfg(**overrides):
    cfg = dict(
        image_size=128,
        in_channels=3,
        model_channels=256,
        out_channels=1000,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 3, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=4,
        num_head_channels=-1,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
        pool="adaptive",
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.classifier.256C")
def adm_classifier_256_cfg(**overrides):
    cfg = dict(
        image_size=256,
        in_channels=3,
        model_channels=256,
        out_channels=1000,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
        pool="adaptive",
    )
    cfg.update(overrides)
    return cfg

@register_model("adm.classifier.512C")
def adm_classifier_512_cfg(**overrides):
    cfg = dict(
        image_size=512,
        in_channels=3,
        model_channels=256,
        out_channels=1000,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        dropout=0.0,
        channel_mult=(0.5, 1, 1, 2, 2, 4, 4),
        conv_resample=True,
        dims=2,
        use_checkpoint=False,
        use_fp16=False,
        num_heads=-1,
        num_head_channels=64,
        num_heads_upsample=-1,
        use_scale_shift_norm=True,
        resblock_updown=True,
        use_new_attention_order=False,
        pool="adaptive",
    )
    cfg.update(overrides)
    return cfg


class ADM_U(ADM):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class ADM_G(torch.nn.Module):
    """
    This model is two stage without possibility of end-to-end training! Only for inference/testing.
    """
    def __init__(self, diffmodel_cfg, classmodel_cfg):
        super().__init__()
        self.diffusion = instantiate_from_config(diffmodel_cfg).eval()
        self.classifier = instantiate_from_config(classmodel_cfg).eval()

    def cond_fn(self, x, t, y=None, classifier_scale=1.):
        assert y is not None
        with torch.enable_grad():
            x_in = x.detach().requires_grad_(True)
            logits = self.classifier(x_in, t)
            log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
            selected = log_probs[range(len(logits)), y.view(-1)]
            return torch.autograd.grad(selected.sum(), x_in)[0] * classifier_scale

    def forward(self, *args, **kwargs): 
        return self.diffusion.forward(*args, **kwargs)


from .detok import *  # noqa: F401,F403
from .maetok import *  # noqa: F401,F403
from .titok import *  # noqa: F401,F403
from .vita import *  # noqa: F401,F403
from .vmae import *  # noqa: F401,F403
from .register import *  # noqa: F401,F403

__all__: list[str] = []


"""
DCAE (Deep Convolutional Autoencoder) modules following the same pattern as ldm_modules.py
Adopted from EfficientViT's DCAE implementation
"""
import torch
import torch.nn as nn
import numpy as np
from typing import Optional, Union, List, Tuple

# Import DCAE building blocks
try:
    from .nn.act import build_act
    from .nn.norm import build_norm
    from .nn.ops import (
        ChannelDuplicatingPixelUnshuffleUpSampleLayer,
        ConvLayer,
        ConvPixelShuffleUpSampleLayer,
        ConvPixelUnshuffleDownSampleLayer,
        EfficientViTBlock,
        IdentityLayer,
        InterpolateConvUpSampleLayer,
        OpSequential,
        PixelUnshuffleChannelAveragingDownSampleLayer,
        ResBlock,
        ResidualBlock,
    )
except ImportError:
    raise ImportError(
        "DCAE modules require the dcae package. Please ensure dcae is installed and in your path."
    )


def build_block(
    block_type: str,
    in_channels: int,
    out_channels: int,
    norm: Optional[str],
    act: Optional[str],
    dims: int = 2,
) -> nn.Module:
    """Build a single block (ResBlock or EfficientViTBlock). EfficientViT blocks are 2D only.

    EViT_GLU / EViTS5_GLU use EfficientViTBlock: LiteMLA (lightweight linear attention over
    spatial dim, with Q,K,V from convs) + GLUMBConv (convolutions with GLU). Not a
    standard Vision Transformer backbone; see EfficientViTBlock docstring in nn.ops.
    """
    if block_type == "ResBlock":
        assert in_channels == out_channels
        main_block = ResBlock(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            use_bias=(True, False),
            norm=(None, norm),
            act_func=(act, None),
            dims=dims,
        )
        block = ResidualBlock(main_block, IdentityLayer())
    elif block_type == "EViT_GLU":
        if dims != 2:
            raise ValueError("EfficientViT blocks (EViT_GLU, EViTS5_GLU) are only supported for 2D (dims=2). Use ResBlock for 3D.")
        assert in_channels == out_channels
        block = EfficientViTBlock(in_channels, norm=norm, act_func=act, local_module="GLUMBConv", scales=())
    elif block_type == "EViTS5_GLU":
        if dims != 2:
            raise ValueError("EfficientViT blocks (EViT_GLU, EViTS5_GLU) are only supported for 2D (dims=2). Use ResBlock for 3D.")
        assert in_channels == out_channels
        block = EfficientViTBlock(in_channels, norm=norm, act_func=act, local_module="GLUMBConv", scales=(5,))
    else:
        raise ValueError(f"block_type {block_type} is not supported")
    return block


def build_stage_main(
    width: int,
    depth: int,
    block_type: Union[str, List[str]],
    norm: str,
    act: str,
    input_width: int,
    dims: int = 2,
) -> List[nn.Module]:
    """Build the main blocks for a stage."""
    assert isinstance(block_type, str) or (isinstance(block_type, list) and depth == len(block_type))
    stage = []
    for d in range(depth):
        current_block_type = block_type[d] if isinstance(block_type, list) else block_type
        block = build_block(
            block_type=current_block_type,
            in_channels=width if d > 0 else input_width,
            out_channels=width,
            norm=norm,
            act=act,
            dims=dims,
        )
        stage.append(block)
    return stage


def build_downsample_block(
    block_type: str,
    in_channels: int,
    out_channels: int,
    shortcut: Optional[str],
    dims: int = 2,
) -> nn.Module:
    """Build a downsampling block."""
    if block_type == "Conv":
        block = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=2,
            use_bias=True,
            norm=None,
            act_func=None,
            dims=dims,
        )
    elif block_type == "ConvPixelUnshuffle":
        block = ConvPixelUnshuffleDownSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2, dims=dims
        )
    else:
        raise ValueError(f"block_type {block_type} is not supported for downsampling")

    if shortcut is None:
        pass
    elif shortcut == "averaging":
        shortcut_block = PixelUnshuffleChannelAveragingDownSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=2, dims=dims
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for downsample")
    return block


def build_upsample_block(
    block_type: str,
    in_channels: int,
    out_channels: int,
    shortcut: Optional[str],
    dims: int = 2,
) -> nn.Module:
    """Build an upsampling block."""
    if block_type == "ConvPixelShuffle":
        block = ConvPixelShuffleUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2, dims=dims
        )
    elif block_type == "InterpolateConv":
        block = InterpolateConvUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2, dims=dims
        )
    else:
        raise ValueError(f"block_type {block_type} is not supported for upsampling")

    if shortcut is None:
        pass
    elif shortcut == "duplicating":
        shortcut_block = ChannelDuplicatingPixelUnshuffleUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=2, dims=dims
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for upsample")
    return block


class DCAEEncoder(nn.Module):
    """
    DCAE Encoder following the same pattern as ldm_modules.Encoder
    
    Args:
        in_channels: Number of input channels
        z_channels: Number of latent channels (output channels)
        width_list: List of channel widths for each stage
        depth_list: List of depths (number of blocks) for each stage
        block_type: Block type string or list of block types per stage
        norm: Normalization type (e.g., "trms2d", "bn2d")
        act: Activation type (e.g., "silu", "relu")
        downsample_block_type: Type of downsampling block ("Conv" or "ConvPixelUnshuffle")
        downsample_match_channel: Whether to match channels in downsampling
        downsample_shortcut: Shortcut type for downsampling ("averaging" or None)
        out_norm: Output normalization type
        out_act: Output activation type
        out_shortcut: Output shortcut type ("averaging" or None)
        double_z: Whether to output double channels (for VAE)
        dims: Number of spatial dimensions (2 or 3)
        img_size: Input image size (for computing vae_stride)
    """
    def __init__(
        self,
        *,
        in_channels=3,
        z_channels=32,
        width_list=(128, 256, 512, 512, 1024, 1024),
        depth_list=(2, 2, 2, 2, 2, 2),
        block_type="ResBlock",
        norm="trms2d",
        act="silu",
        downsample_block_type="ConvPixelUnshuffle",
        downsample_match_channel=True,
        downsample_shortcut="averaging",
        out_norm=None,
        out_act=None,
        out_shortcut="averaging",
        project_out_conv_only=False,
        double_z=False,
        dims=2,
        img_size=256,
        **ignore_kwargs,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.z_channels = z_channels
        self.dims = dims
        self.num_stages = len(width_list)
        self.vae_stride = 2 ** (self.num_stages - 1)
        
        assert len(depth_list) == self.num_stages
        assert len(width_list) == self.num_stages
        assert isinstance(block_type, str) or (
            isinstance(block_type, list) and len(block_type) == self.num_stages
        )

        # Project in
        project_in_factor = 1 if depth_list[0] > 0 else 2
        project_in_out_channels = width_list[0] if depth_list[0] > 0 else width_list[1]
        
        if project_in_factor == 1:
            self.project_in = ConvLayer(
                in_channels=in_channels,
                out_channels=project_in_out_channels,
                kernel_size=3,
                stride=1,
                use_bias=True,
                norm=None,
                act_func=None,
                dims=self.dims,
            )
        elif project_in_factor == 2:
            self.project_in = build_downsample_block(
                block_type=downsample_block_type,
                in_channels=in_channels,
                out_channels=project_in_out_channels,
                shortcut=None,
                dims=self.dims,
            )
        else:
            raise ValueError(f"downsample factor {project_in_factor} is not supported for encoder project in")

        # Build stages
        self.stages = nn.ModuleList()
        for stage_id, (width, depth) in enumerate(zip(width_list, depth_list)):
            current_block_type = block_type[stage_id] if isinstance(block_type, list) else block_type
            stage_blocks = build_stage_main(
                width=width,
                depth=depth,
                block_type=current_block_type,
                norm=norm,
                act=act,
                input_width=width,
                dims=self.dims,
            )

            if stage_id < self.num_stages - 1 and depth > 0:
                downsample_block = build_downsample_block(
                    block_type=downsample_block_type,
                    in_channels=width,
                    out_channels=width_list[stage_id + 1] if downsample_match_channel else width,
                    shortcut=downsample_shortcut,
                    dims=self.dims,
                )
                stage_blocks.append(downsample_block)

            self.stages.append(OpSequential(stage_blocks))

        # Project out
        out_channels = 2 * z_channels if double_z else z_channels
        if project_out_conv_only:
            # Single conv at op_list.0 to match pretrained DCAE checkpoint keys
            project_out_layers = [
                ConvLayer(
                    in_channels=width_list[-1],
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=1,
                    use_bias=True,
                    norm=None,
                    act_func=None,
                    dims=self.dims,
                ),
            ]
        else:
            project_out_layers = [
                build_norm(out_norm, num_features=width_list[-1], dims=self.dims) if out_norm else IdentityLayer(),
                build_act(out_act) if out_act else IdentityLayer(),
                ConvLayer(
                    in_channels=width_list[-1],
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=1,
                    use_bias=True,
                    norm=None,
                    act_func=None,
                    dims=self.dims,
                ),
            ]

        if out_shortcut == "averaging":
            shortcut_block = PixelUnshuffleChannelAveragingDownSampleLayer(
                in_channels=width_list[-1], out_channels=out_channels, factor=1, dims=self.dims
            )
            self.project_out = ResidualBlock(OpSequential(project_out_layers), shortcut_block)
        else:
            self.project_out = OpSequential(project_out_layers)

    def forward(self, x):
        x = self.project_in(x)
        for stage in self.stages:
            if len(stage.op_list) == 0:
                continue
            x = stage(x)
        x = self.project_out(x)
        return x


class DCAEDecoder(nn.Module):
    """
    DCAE Decoder following the same pattern as ldm_modules.Decoder
    
    Args:
        in_channels: Number of input channels (for output)
        z_channels: Number of latent channels (input channels)
        width_list: List of channel widths for each stage
        depth_list: List of depths (number of blocks) for each stage
        block_type: Block type string or list of block types per stage
        norm: Normalization type (string or list per stage)
        act: Activation type (string or list per stage)
        upsample_block_type: Type of upsampling block ("ConvPixelShuffle" or "InterpolateConv")
        upsample_match_channel: Whether to match channels in upsampling
        upsample_shortcut: Shortcut type for upsampling ("duplicating" or None)
        in_shortcut: Input shortcut type ("duplicating" or None)
        out_norm: Output normalization type
        out_act: Output activation type
        dims: Number of spatial dimensions (2 or 3)
        img_size: Input image size (for computing z_shape)
    """
    def __init__(
        self,
        *,
        in_channels=3,
        z_channels=32,
        width_list=(128, 256, 512, 512, 1024, 1024),
        depth_list=(2, 2, 2, 2, 2, 2),
        block_type="ResBlock",
        norm="trms2d",
        act="silu",
        upsample_block_type="ConvPixelShuffle",
        upsample_match_channel=True,
        upsample_shortcut="duplicating",
        in_shortcut="duplicating",
        out_norm="trms2d",
        out_act="relu",
        dims=2,
        img_size=256,
        **ignore_kwargs,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.z_channels = z_channels
        self.dims = dims
        self.num_stages = len(width_list)
        
        assert len(depth_list) == self.num_stages
        assert len(width_list) == self.num_stages
        assert isinstance(block_type, str) or (
            isinstance(block_type, list) and len(block_type) == self.num_stages
        )
        assert isinstance(norm, str) or (isinstance(norm, list) and len(norm) == self.num_stages)
        assert isinstance(act, str) or (isinstance(act, list) and len(act) == self.num_stages)

        # Compute z_shape
        curr_res = img_size // 2 ** (self.num_stages - 1)
        if dims == 2:
            self.z_shape = (1, z_channels, curr_res, curr_res)
        else:  # dims == 3
            self.z_shape = (1, z_channels, curr_res, curr_res, curr_res)
        
        print(
            "Working with z of shape {} = {} dimensions.".format(
                self.z_shape, np.prod(self.z_shape)
            )
        )

        # Project in
        project_in_block = ConvLayer(
            in_channels=z_channels,
            out_channels=width_list[-1],
            kernel_size=3,
            stride=1,
            use_bias=True,
            norm=None,
            act_func=None,
            dims=self.dims,
        )

        if in_shortcut == "duplicating":
            shortcut_block = ChannelDuplicatingPixelUnshuffleUpSampleLayer(
                in_channels=z_channels, out_channels=width_list[-1], factor=1, dims=self.dims
            )
            self.project_in = ResidualBlock(project_in_block, shortcut_block)
        else:
            self.project_in = project_in_block

        # Build stages (in reverse order)
        self.stages = []
        for stage_id, (width, depth) in reversed(list(enumerate(zip(width_list, depth_list)))):
            stage_blocks = []
            
            if stage_id < self.num_stages - 1 and depth > 0:
                upsample_block = build_upsample_block(
                    block_type=upsample_block_type,
                    in_channels=width_list[stage_id + 1],
                    out_channels=width if upsample_match_channel else width_list[stage_id + 1],
                    shortcut=upsample_shortcut,
                    dims=self.dims,
                )
                stage_blocks.append(upsample_block)

            current_block_type = block_type[stage_id] if isinstance(block_type, list) else block_type
            current_norm = norm[stage_id] if isinstance(norm, list) else norm
            current_act = act[stage_id] if isinstance(act, list) else act

            stage_blocks.extend(
                build_stage_main(
                    width=width,
                    depth=depth,
                    block_type=current_block_type,
                    norm=current_norm,
                    act=current_act,
                    input_width=(
                        width if upsample_match_channel else width_list[min(stage_id + 1, self.num_stages - 1)]
                    ),
                    dims=self.dims,
                )
            )
            self.stages.insert(0, OpSequential(stage_blocks))
        
        self.stages = nn.ModuleList(self.stages)

        # Project out
        project_out_factor = 1 if depth_list[0] > 0 else 2
        project_out_in_channels = width_list[0] if depth_list[0] > 0 else width_list[1]
        
        project_out_layers = [
            build_norm(out_norm, num_features=project_out_in_channels, dims=self.dims) if out_norm else IdentityLayer(),
            build_act(out_act) if out_act else IdentityLayer(),
        ]

        if project_out_factor == 1:
            project_out_layers.append(
                ConvLayer(
                    in_channels=project_out_in_channels,
                    out_channels=in_channels,
                    kernel_size=3,
                    stride=1,
                    use_bias=True,
                    norm=None,
                    act_func=None,
                    dims=self.dims,
                )
            )
        elif project_out_factor == 2:
            project_out_layers.append(
                build_upsample_block(
                    block_type=upsample_block_type,
                    in_channels=project_out_in_channels,
                    out_channels=in_channels,
                    shortcut=None,
                    dims=self.dims,
                )
            )
        else:
            raise ValueError(f"upsample factor {project_out_factor} is not supported for decoder project out")
        
        self.project_out = OpSequential(project_out_layers)

    def forward(self, z):
        x = self.project_in(z)
        for stage in reversed(self.stages):
            if len(stage.op_list) == 0:
                continue
            x = stage(x)
        x = self.project_out(x)
        return x

    def get_last_layer(self) -> nn.Parameter:
        """Get the last layer weights (for loss computation)."""
        return self.project_out.op_list[-1].conv.weight if hasattr(self.project_out.op_list[-1], 'conv') else None



from typing import Any, Optional

import numpy as np
import torch
import torch.nn as nn

from .ops import IdentityLayer, ResidualBlock
from ..utils.network import build_kwargs_from_config

# Scheduler placeholder - you may need to implement this or import from elsewhere
class Scheduler:
    PROGRESS = 1.0

__all__ = ["apply_drop_func"]


def apply_drop_func(network: nn.Module, drop_config: Optional[dict[str, Any]]) -> None:
    if drop_config is None:
        return

    drop_lookup_table = {
        "droppath": apply_droppath,
    }

    drop_func = drop_lookup_table[drop_config["name"]]
    drop_kwargs = build_kwargs_from_config(drop_config, drop_func)

    drop_func(network, **drop_kwargs)


def apply_droppath(
    network: nn.Module,
    drop_prob: float,
    linear_decay=True,
    scheduled=True,
    skip=0,
) -> None:
    all_valid_blocks = []
    for m in network.modules():
        for name, sub_module in m.named_children():
            if isinstance(sub_module, ResidualBlock) and isinstance(sub_module.shortcut, IdentityLayer):
                all_valid_blocks.append((m, name, sub_module))
    all_valid_blocks = all_valid_blocks[skip:]
    for i, (m, name, sub_module) in enumerate(all_valid_blocks):
        prob = drop_prob * (i + 1) / len(all_valid_blocks) if linear_decay else drop_prob
        new_module = DropPathResidualBlock(
            sub_module.main,
            sub_module.shortcut,
            sub_module.post_act,
            sub_module.pre_norm,
            prob,
            scheduled,
        )
        m._modules[name] = new_module


class DropPathResidualBlock(ResidualBlock):
    def __init__(
        self,
        main: nn.Module,
        shortcut: Optional[nn.Module],
        post_act=None,
        pre_norm: Optional[nn.Module] = None,
        ######################################
        drop_prob: float = 0,
        scheduled=True,
    ):
        super().__init__(main, shortcut, post_act, pre_norm)

        self.drop_prob = drop_prob
        self.scheduled = scheduled

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.drop_prob == 0 or not isinstance(self.shortcut, IdentityLayer):
            return ResidualBlock.forward(self, x)
        else:
            drop_prob = self.drop_prob
            if self.scheduled:
                drop_prob *= np.clip(Scheduler.PROGRESS, 0, 1)
            keep_prob = 1 - drop_prob

            shape = (x.shape[0],) + (1,) * (x.ndim - 1)
            random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
            random_tensor.floor_()  # binarize

            res = self.forward_main(x) / keep_prob * random_tensor + self.shortcut(x)
            if self.post_act:
                res = self.post_act(res)
            return res



import torch 
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from einops import rearrange, pack, unpack
from typing import List, Tuple, Optional, Sequence, Union
from math import ceil

def round_up_multiple(num, mult):
    return ceil(num / mult) * mult


class EmbeddingEMA(nn.Module):
    """
    EMA-based embedding that can be used as a drop-in replacement for nn.Embedding.
    Provides exponential moving average updates for codebook embeddings during training.

    Args:
            num_tokens: Number of tokens in codebook
            codebook_dim: Dimension of embedding
            decay: EMA decay rate
            eps: Epsilon value for numerical stability
    """
    def __init__(self, num_tokens, codebook_dim, decay=0.99, eps=1e-5):

        super().__init__()
        self.decay = decay
        self.eps = eps        
        weight = torch.randn(num_tokens, codebook_dim)
        self.weight = nn.Parameter(weight, requires_grad = False)
        self.cluster_size = nn.Parameter(torch.zeros(num_tokens), requires_grad = False)
        self.embed_avg = nn.Parameter(weight.clone(), requires_grad = False)
        self.update = True

    def forward(self, embed_id):
        """Drop-in replacement for nn.Embedding.forward()"""
        return F.embedding(embed_id, self.weight)

    def cluster_size_ema_update(self, new_cluster_size):
        """Update cluster size with EMA."""
        self.cluster_size.data.mul_(self.decay).add_(new_cluster_size, alpha=1 - self.decay)

    def embed_avg_ema_update(self, new_embed_avg): 
        """Update embedding average with EMA."""
        self.embed_avg.data.mul_(self.decay).add_(new_embed_avg, alpha=1 - self.decay)

    def weight_update(self, num_tokens):
        """Update embedding weights based on EMA statistics."""
        n = self.cluster_size.sum()
        smoothed_cluster_size = (
                (self.cluster_size + self.eps) / (n + num_tokens * self.eps) * n
            )
        #normalize embedding average with smoothed cluster size
        embed_normalized = self.embed_avg / smoothed_cluster_size.unsqueeze(1)
        self.weight.data.copy_(embed_normalized)
    
    def perform_ema_update(self, encodings, z_flattened, num_tokens):
        """
        Perform EMA update given encodings and flattened input features.
        
        Args:
            encodings: One-hot encodings of shape (N, num_tokens) where N is flattened spatial dims
            z_flattened: Flattened input features of shape (N, codebook_dim)
            num_tokens: Total number of tokens in codebook
        """
        if self.training and self.update:
            # EMA cluster size
            encodings_sum = encodings.sum(0)
            self.cluster_size_ema_update(encodings_sum)
            # EMA embedding average
            embed_sum = encodings.transpose(0, 1) @ z_flattened
            self.embed_avg_ema_update(embed_sum)
            # Normalize embed_avg and update weight
            self.weight_update(num_tokens)   

class QincoSubstep(nn.Module):
    def __init__(self, e_dim: int, hidden_dim: int = 256, num_layers: int = 3):
        super().__init__()
        in_dim = e_dim * 2  # concat(base_code, x_prev)
        layers = []
        for _ in range(num_layers - 1):
            layers.append(nn.Linear(in_dim, hidden_dim))
            layers.append(nn.ReLU())
            in_dim = hidden_dim
        layers.append(nn.Linear(hidden_dim, e_dim))
        self.net = nn.Sequential(*layers)

    def forward(self, base_code: torch.Tensor, x_prev: torch.Tensor) -> torch.Tensor:
        """
        base_code: (B, D)
        x_prev:    (B, D)
        returns:   (B, D) delta
        """
        inp = torch.cat([base_code, x_prev], dim=-1)
        return self.net(inp)

class ImplicitEmbedding(nn.Module):
    """
    Implicit embedding that can be used as a drop-in replacement for nn.Embedding.
    Provides a small MLP to map coordinates to codebook vectors.
    QINCode: https://arxiv.org/abs/2401.14732

    Args:
        n_e: Number of codes in codebook
        dim: Dimension of embedding
        hidden_dim: Hidden dimension of MLP
        num_layers: Number of layers of MLP
    """
    def __init__(self, n_e, e_dim, hidden_dim=256, num_layers=3):
        super().__init__()
        self.n_e = n_e
        self.e_dim = e_dim

        # coords or latent code parameters
        self.coords = nn.Parameter(torch.randn(n_e, e_dim))

        # small MLP: coords → code vectors
        layers = []
        in_dim = e_dim
        for _ in range(num_layers - 1):
            layers.append(nn.Linear(in_dim, hidden_dim))
            layers.append(nn.ReLU())
            in_dim = hidden_dim
        layers.append(nn.Linear(hidden_dim, e_dim))
        self.mlp = nn.Sequential(*layers)

    @property
    def weight(self):
        """Acts like nn.Embedding.weight: returns full codebook matrix."""
        return self.mlp(self.coords)   # (n_e, e_dim)

    def forward(self, indices):
        """Acts like nn.Embedding(indices)."""
        codes = self.weight                 # (n_e, e_dim)
        return codes[indices]               # indexing works identically


class Phi(nn.Conv2d):
    def __init__(self, embed_dim, quant_resi):
        ks = 3
        super().__init__(in_channels=embed_dim, out_channels=embed_dim, kernel_size=ks, stride=1, padding=ks//2)
        self.resi_ratio = abs(quant_resi)
    
    def forward(self, h_BChw):
        return h_BChw.mul(1-self.resi_ratio) + super().forward(h_BChw).mul_(self.resi_ratio)


class Phi3D(nn.Conv3d):
    def __init__(self, embed_dim, quant_resi):
        ks = 3
        super().__init__(in_channels=embed_dim, out_channels=embed_dim, kernel_size=ks, stride=1, padding=ks//2)
        self.resi_ratio = abs(quant_resi)
    
    def forward(self, h_BCdhw):
        return h_BCdhw.mul(1-self.resi_ratio) + super().forward(h_BCdhw).mul_(self.resi_ratio)


class PhiShared(nn.Module):
    def __init__(self, qresi: Phi):
        super().__init__()
        self.qresi: Phi = qresi
    
    def __getitem__(self, _) -> Phi:
        return self.qresi


class PhiPartiallyShared(nn.Module):
    def __init__(self, qresi_ls: nn.ModuleList):
        super().__init__()
        self.qresi_ls = qresi_ls
        K = len(qresi_ls)
        self.ticks = np.linspace(1/3/K, 1-1/3/K, K) if K == 4 else np.linspace(1/2/K, 1-1/2/K, K)
    
    def __getitem__(self, at_from_0_to_1: float) -> Phi:
        return self.qresi_ls[np.argmin(np.abs(self.ticks - at_from_0_to_1)).item()]
    
    def extra_repr(self) -> str:
        return f'ticks={self.ticks}'


class PhiNonShared(nn.ModuleList):
    def __init__(self, qresi: List):
        super().__init__(qresi)
        # self.qresi = qresi
        K = len(qresi)
        self.ticks = np.linspace(1/3/K, 1-1/3/K, K) if K == 4 else np.linspace(1/2/K, 1-1/2/K, K)
    
    def __getitem__(self, at_from_0_to_1: float) -> Phi:
        return super().__getitem__(np.argmin(np.abs(self.ticks - at_from_0_to_1)).item())
    
    def extra_repr(self) -> str:
        return f'ticks={self.ticks}'
    
def safe_div(num, den, eps = 1e-6):
    return num / den.clamp(min = eps)

def l2norm(t, dim = -1,  eps = 1e-6):
    return F.normalize(t, p = 2, dim = dim, eps = eps)

def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d

def pack_one(t, pattern):
    packed, ps = pack([t], pattern)

    def unpack_one(to_unpack, unpack_pattern = None):
        unpacked, = unpack(to_unpack, ps, default(unpack_pattern, pattern))
        return unpacked

    return packed, unpack_one

def efficient_rotation_trick_transform(u, q, e):
    """
    4.2 in https://arxiv.org/abs/2410.06424
    """
    e = rearrange(e, 'b d -> b 1 d')
    w = l2norm(u + q, dim = 1).detach()

    return (
        e -
        2 * (e @ rearrange(w, 'b d -> b d 1') @ rearrange(w, 'b d -> b 1 d')) +
        2 * (e @ rearrange(u, 'b d -> b d 1').detach() @ rearrange(q, 'b d -> b 1 d').detach())
    )

def rotate_to(src, tgt):
    # rotation trick STE (https://arxiv.org/abs/2410.06424) to get gradients through VQ layer.
    src, inverse = pack_one(src, '* d')
    tgt, _ = pack_one(tgt, '* d')

    norm_src = src.norm(dim = -1, keepdim = True)
    norm_tgt = tgt.norm(dim = -1, keepdim = True)

    rotated_tgt = efficient_rotation_trick_transform(
        safe_div(src, norm_src),
        safe_div(tgt, norm_tgt),
        src
    ).squeeze()

    rotated = rotated_tgt * safe_div(norm_tgt, norm_src).detach()

    return inverse(rotated)

def clamp_log(x: torch.Tensor, eps: float = 1e-5) -> torch.Tensor:
    """ Clamps the input tensor and computes the log.

    Args:
        x -> torch.Tensor: The input tensor.
        eps -> float: The epsilon value serving as the lower bound.

    Returns:
        torch.Tensor: The log of the clamped input tensor.
    """
    return torch.log(torch.clamp(x, eps))

def entropy_loss_fn(
    affinity: torch.Tensor,
    temperature: float,
    entropy_gamma: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """ Computes the entropy loss.

    Args:
        affinity -> torch.Tensor: The affinity matrix.
        temperature -> float: The temperature.
        entropy_gamma -> float: The entropy gamma.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The per-sample and average entropy.
    """
    flat_affinity = affinity.reshape(-1, affinity.shape[-1])
    flat_affinity /= temperature

    probability = flat_affinity.softmax(dim=-1)
    average_probability = torch.mean(probability, dim=0)

    per_sample_entropy = -1 * torch.mean(torch.sum(probability * clamp_log(probability), dim=-1))
    avg_entropy = torch.sum(-1 * average_probability * clamp_log(average_probability))

    return (per_sample_entropy, avg_entropy * entropy_gamma)
